###
# Copyright CEA (Commissariat à l'énergie atomique et aux
# énergies alternatives) (2017-2025)
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
###

import ctypes
import logging
import math
from collections import OrderedDict
from typing import Iterable, Mapping, Tuple

from secbench.api import Discoverable, HardwareInfo
from secbench.api.enums import Slope
from secbench.api.instrument import ScopeAnalogChannel
from secbench.picoscope.lib.ps6000 import PicoInfo, load_ps6000

from .base import Picobase, PicobaseAnalogChannel, PicoHandles
from .error import PicoscopeApiError, pico_check
from .types import VerticalRange

logger = logging.getLogger(__name__)


class PicoPS6000Scope(Discoverable, Picobase):
    _PICO_ALL_CHANNELS = {
        "A": 0,
        "B": 1,
        "C": 2,
        "D": 3,
        "External": 4,
        "TriggerAux": 5,
    }
    _PICO_CHANNELS = {"A": 0, "B": 1, "C": 2, "D": 3}
    _PICO_VERTICAL_RANGES = [
        # Does not seems supported.
        # VerticalRange(20e-3, 1, "20 mV"),
        VerticalRange(50e-3, 2, "50 mV"),
        VerticalRange(100e-3, 3, "100 mV"),
        VerticalRange(200e-3, 4, "200 mV"),
        VerticalRange(500e-3, 5, "500 mV"),
        VerticalRange(1.0, 6, "1 V"),
        VerticalRange(2.0, 7, "2 V"),
        VerticalRange(5.0, 8, "5 V"),
        VerticalRange(10.0, 9, "10 V"),
        VerticalRange(20.0, 10, "20 V"),
        VerticalRange(50.0, 11, "50 V"),
    ]
    _PICO_SLOPES = {
        Slope.rising: 2,
        Slope.falling: 3,
        Slope.either: 4,
    }
    _MAX_SAMPLES = {
        "6402C": 256 * 1024 * 1024,
        "6402D": 512 * 1024 * 1024,
        "6403C": 512 * 1024 * 1024,
        "6403D": 1024 * 1024 * 1024,
        "6404C": 1024 * 1024 * 1024,
        "6404D": 2 * 1024 * 1024 * 1024,
    }

    @staticmethod
    def ps6000_methods(lib):
        handles = PicoHandles(
            psOpenUnit=lib.ps6000OpenUnit,
            psCloseUnit=lib.ps6000CloseUnit,
            psIsReady=lib.ps6000IsReady,
            psGetUnitInfo=lib.ps6000GetUnitInfo,
            psGetTimebase2=lib.ps6000GetTimebase2,
            psStop=lib.ps6000Stop,
            psSetChannel=lib.ps6000SetChannel,
            psSetNoOfCaptures=lib.ps6000SetNoOfCaptures,
            psRunBlock=lib.ps6000RunBlock,
            psSetDataBuffer=lib.ps6000SetDataBuffer,
            psSetDataBufferBulk=lib.ps6000SetDataBufferBulk,
            psGetValues=lib.ps6000GetValues,
            psGetValuesBulk=lib.ps6000GetValuesBulk,
            psSetSimpleTrigger=lib.ps6000SetSimpleTrigger,
            psMemorySegments=lib.ps6000MemorySegments,
            psPingUnit=lib.ps6000PingUnit,
        )
        return handles

    def __init__(self, serial_number=None):
        lib = load_ps6000()
        methods = self.ps6000_methods(lib)
        super().__init__(
            lib, methods, device_info_cls=PicoInfo, serial_number=serial_number
        )
        self._channels = OrderedDict(
            [
                ("A", PicobaseAnalogChannel(self, "A")),
                ("B", PicobaseAnalogChannel(self, "B")),
                ("C", PicobaseAnalogChannel(self, "C")),
                ("D", PicobaseAnalogChannel(self, "D")),
            ]
        )

    # ===
    # Picobase interface
    # ===
    def _pico_adc_range(self) -> Tuple[int, int]:
        return -32_512, 32_512

    @staticmethod
    def _pico_interval_to_timebase(interval: float) -> int:
        max_sample_time = ((2**32 - 1) - 4) / 156_250_000

        if interval < 6.4e-9:
            timebase = math.floor(math.log(interval * 5e9, 2))
            timebase = max(timebase, 0)
        else:
            # Otherwise in range 2^32-1
            if interval > max_sample_time:
                interval = max_sample_time
            timebase = math.floor((interval * 156_250_000) + 4)
        return int(timebase)

    @staticmethod
    def _pico_timebase_to_interval(timebase: int) -> float:
        assert isinstance(timebase, int), "timebase is an integer"
        if 0 <= timebase <= 4:
            return (1 << timebase) / 5e9
        elif 5 <= timebase <= (1 << 32) - 1:
            return (timebase - 4) / 156_250_000
        else:
            # FIXME(@TH): use a more specific error.
            raise ValueError("invalid timebase: {}".format(timebase))

    # ===
    # Discoverable Interface
    # ===
    @classmethod
    def is_supported(cls, hardware_info: HardwareInfo) -> bool:
        try:
            _ = load_ps6000()
            return True
        except Exception as e:
            logger.debug(
                f"unable to load libps6000, PS2000A devices cannot be discovered (exception is {e})"
            )
        return False

    @classmethod
    def discover(cls, hw_info: HardwareInfo) -> Iterable[str]:
        try:
            lib = load_ps6000()
            count = ctypes.c_int16(0)
            serial_len = ctypes.c_int16(4096)
            serial_buff = (ctypes.c_int8 * serial_len.value)()
            pico_check(
                lib.ps6000EnumerateUnits(
                    ctypes.byref(count), serial_buff, ctypes.byref(serial_len)
                )
            )

            serial_ids = bytes(serial_buff[: serial_len.value - 1]).decode()
            logger.debug(f"picoscope devices available: {serial_ids}")
            for serial_num in serial_ids.split(","):
                yield serial_num
        except PicoscopeApiError as e:
            logger.info(f"no ps6000 device found: {e}, leaving...")

    @classmethod
    def build(cls, hardware_info: HardwareInfo, serial_number: str):
        return cls(serial_number=None)

    # ===
    # Scope interface
    # ===

    @property
    def description(self) -> str:
        info = self.device_info()
        variant = info.get(PicoInfo.VARIANT_INFO, "?")
        serial = info.get(PicoInfo.BATCH_AND_SERIAL, "?")
        return f"picoscope-PS6000[variant={variant}, serial={serial}]"

    def channels(self) -> Mapping[str, ScopeAnalogChannel]:
        return self._channels


def def_symbol(lib, name, return_type, arg_types, docstring=None):
    c_function = getattr(lib, name)
    c_function.restype = return_type
    c_function.argtypes = arg_types
    if docstring is not None:
        c_function.__doc__ = docstring


def load_libps(name: str):
    """
    Try to find picoscope library on the current system.

    :param name: prefix of the library (e.g., "ps2000a", "ps6000")
    :returns: a handle (ctypes or windll) on the library.
    """
    import platform

    system = platform.system()
    if system == "Linux":
        from ctypes import cdll

        lib = cdll.LoadLibrary(f"lib{name}.so.2")
    elif system == "Darwin":
        raise NotImplementedError("MacOSX is not supported yet")
    elif system == "Windows":
        from ctypes import windll
        from ctypes.util import find_library

        lib = windll.LoadLibrary(find_library(f"{name}.dll"))
    else:
        raise NotImplementedError(
            f"Unsupported platform for loading lib{name}: {system}"
        )
    return lib


###
# Copyright CEA (Commissariat à l'énergie atomique et aux
# énergies alternatives) (2017-2025)
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
###

"""
PS2000A driver using direct library calls
"""

from ctypes import (
    c_char_p,
    c_double,
    c_float,
    c_int16,
    c_int32,
    c_uint32,
    c_uint64,
    c_void_p,
)
from enum import IntEnum

from .common import def_symbol, load_libps


class PicoInfo(IntEnum):
    """
    Information to be requested to `ps2000aGetUnitInfo`.
    """

    DRIVER_VERSION = 0x00000000
    USB_VERSION = 0x00000001
    HARDWARE_VERSION = 0x00000002
    VARIANT_INFO = 0x00000003
    BATCH_AND_SERIAL = 0x00000004
    CAL_DATE = 0x00000005
    KERNEL_VERSION = 0x00000006
    DIGITAL_HARDWARE_VERSION = 0x00000007
    ANALOGUE_HARDWARE_VERSION = 0x00000008
    FIRMWARE_VERSION_1 = 0x00000009
    FIRMWARE_VERSION_2 = 0x0000000A


def _load_ps2000a_lib():
    lib = load_libps("ps2000a")
    def_symbol(lib, "ps2000aOpenUnit", c_uint32, [c_void_p, c_char_p])
    def_symbol(lib, "ps2000aOpenUnitAsync", c_uint32, [c_void_p, c_char_p])

    def_symbol(lib, "ps2000aOpenUnitProgress", c_uint32, [c_void_p, c_void_p, c_void_p])

    def_symbol(
        lib,
        "ps2000aGetUnitInfo",
        c_uint32,
        [c_int16, c_void_p, c_int16, c_void_p, c_uint32],
    )

    def_symbol(lib, "ps2000aFlashLed", c_uint32, [c_int16, c_int16])

    def_symbol(
        lib,
        "ps2000aCloseUnit",
        c_uint32,
        [
            c_int16,
        ],
    )

    def_symbol(lib, "ps2000aMemorySegments", c_uint32, [c_int16, c_uint32, c_void_p])

    def_symbol(
        lib,
        "ps2000aSetChannel",
        c_uint32,
        [c_int16, c_int32, c_int16, c_int32, c_int32, c_float],
    )

    def_symbol(
        lib, "ps2000aSetDigitalPort", c_uint32, [c_int16, c_int32, c_int16, c_int16]
    )

    def_symbol(lib, "ps2000aSetNoOfCaptures", c_uint32, [c_int16, c_uint32])
    def_symbol(
        lib,
        "ps2000aGetTimebase",
        c_uint32,
        [c_int16, c_uint32, c_int32, c_void_p, c_int16, c_void_p, c_uint32],
    )
    def_symbol(
        lib,
        "ps2000aGetTimebase2",
        c_uint32,
        [c_int16, c_uint32, c_int32, c_void_p, c_int16, c_void_p, c_uint32],
    )
    def_symbol(
        lib,
        "ps2000aSetSigGenArbitrary",
        c_uint32,
        [
            c_int16,
            c_int32,
            c_uint32,
            c_uint32,
            c_uint32,
            c_uint32,
            c_uint32,
            c_void_p,
            c_int32,
            c_int32,
            c_int32,
            c_int32,
            c_uint32,
            c_uint32,
            c_int32,
            c_int32,
            c_int16,
        ],
    )
    def_symbol(
        lib,
        "ps2000aSetSigGenBuiltIn",
        c_uint32,
        [
            c_int16,
            c_int32,
            c_uint32,
            c_int16,
            c_float,
            c_float,
            c_float,
            c_float,
            c_int32,
            c_int32,
            c_uint32,
            c_uint32,
            c_int32,
            c_int32,
            c_int16,
        ],
    )
    def_symbol(
        lib,
        "ps2000aSetSigGenPropertiesArbitrary",
        c_uint32,
        [
            c_int16,
            c_uint32,
            c_uint32,
            c_uint32,
            c_uint32,
            c_int32,
            c_uint32,
            c_uint32,
            c_int32,
            c_int32,
            c_int16,
        ],
    )
    def_symbol(
        lib,
        "ps2000aSetSigGenPropertiesBuiltIn",
        c_uint32,
        [
            c_int16,
            c_double,
            c_double,
            c_double,
            c_double,
            c_int32,
            c_uint32,
            c_uint32,
            c_int32,
            c_int32,
            c_int16,
        ],
    )
    def_symbol(
        lib,
        "ps2000aSigGenFrequencyToPhase",
        c_uint32,
        [c_int16, c_double, c_int32, c_uint32, c_void_p],
    )
    def_symbol(
        lib,
        "ps2000aSigGenArbitraryMinMaxValues",
        c_uint32,
        [c_int16, c_void_p, c_void_p, c_void_p, c_void_p],
    )
    def_symbol(lib, "ps2000aSigGenSoftwareControl", c_uint32, [c_int16, c_int16])
    def_symbol(
        lib, "ps2000aSetEts", c_uint32, [c_int16, c_int32, c_int16, c_int16, c_void_p]
    )
    def_symbol(
        lib,
        "ps2000aSetSimpleTrigger",
        c_uint32,
        [c_int16, c_int16, c_int32, c_int16, c_int32, c_uint32, c_int16],
    )
    def_symbol(
        lib,
        "ps2000aSetTriggerDigitalPortProperties",
        c_uint32,
        [c_int16, c_void_p, c_int16],
    )
    def_symbol(
        lib,
        "ps2000aSetDigitalAnalogTriggerOperand",
        c_uint32,
        [c_int16, c_int32],
    )
    def_symbol(
        lib,
        "ps2000aSetTriggerChannelProperties",
        c_uint32,
        [c_int16, c_void_p, c_int16, c_int16, c_int32],
    )
    def_symbol(
        lib,
        "ps2000aSetTriggerChannelConditions",
        c_uint32,
        [c_int16, c_void_p, c_int16],
    )
    def_symbol(
        lib,
        "ps2000aSetTriggerChannelDirections",
        c_uint32,
        [c_int16, c_int32, c_int32, c_int32, c_int32, c_int32, c_int32],
    )
    def_symbol(lib, "ps2000aSetTriggerDelay", c_uint32, [c_int16, c_uint32])
    def_symbol(
        lib,
        "ps2000aSetPulseWidthQualifier",
        c_uint32,
        [c_int16, c_void_p, c_int16, c_int32, c_uint32, c_uint32, c_int32],
    )
    def_symbol(
        lib,
        "ps2000aIsTriggerOrPulseWidthQualifierEnabled",
        c_uint32,
        [c_int16, c_void_p, c_void_p],
    )
    def_symbol(
        lib,
        "ps2000aGetTriggerTimeOffset64",
        c_uint32,
        [c_int16, c_void_p, c_void_p, c_uint32],
    )
    def_symbol(
        lib,
        "ps2000aGetValuesTriggerTimeOffsetBulk64",
        c_uint32,
        [c_int16, c_void_p, c_void_p, c_uint32, c_uint32],
    )
    def_symbol(lib, "ps2000aGetNoOfCaptures", c_uint32, [c_int16, c_void_p])
    def_symbol(lib, "ps2000aGetNoOfProcessedCaptures", c_uint32, [c_int16, c_void_p])
    def_symbol(
        lib,
        "ps2000aSetDataBuffer",
        c_uint32,
        [c_int16, c_int32, c_void_p, c_int32, c_uint32, c_int32],
    )
    def_symbol(
        lib,
        "ps2000aSetDataBuffers",
        c_uint32,
        [c_int16, c_int32, c_void_p, c_void_p, c_int32, c_uint32, c_int32],
    )
    def_symbol(lib, "ps2000aSetEtsTimeBuffer", c_uint32, [c_int16, c_void_p, c_int32])
    def_symbol(lib, "ps2000aIsReady", c_uint32, [c_int16, c_void_p])
    def_symbol(
        lib,
        "ps2000aRunBlock",
        c_uint32,
        [
            c_int16,
            c_int32,
            c_int32,
            c_uint32,
            c_int16,
            c_void_p,
            c_uint32,
            c_void_p,
            c_void_p,
        ],
    )
    def_symbol(
        lib,
        "ps2000aRunStreaming",
        c_uint32,
        [
            c_int16,
            c_void_p,
            c_int32,
            c_uint32,
            c_uint32,
            c_int16,
            c_uint32,
            c_int32,
            c_uint32,
        ],
    )
    def_symbol(
        lib, "ps2000aGetStreamingLatestValues", c_uint32, [c_int16, c_void_p, c_void_p]
    )
    def_symbol(
        lib,
        "ps2000aGetMaxDownSampleRatio",
        c_uint32,
        [c_int16, c_uint32, c_void_p, c_int32, c_uint32],
    )
    def_symbol(
        lib,
        "ps2000aGetValues",
        c_uint32,
        [c_int16, c_uint32, c_void_p, c_uint32, c_int32, c_uint32, c_void_p],
    )

    def_symbol(
        lib,
        "ps2000aGetValuesBulk",
        c_uint32,
        [c_int16, c_void_p, c_uint32, c_uint32, c_uint32, c_int32, c_void_p],
    )
    def_symbol(
        lib,
        "ps2000aGetValuesAsync",
        c_uint32,
        [c_int16, c_uint32, c_uint32, c_uint32, c_int32, c_uint32, c_void_p, c_void_p],
    )
    def_symbol(
        lib,
        "ps2000aGetValuesOverlapped",
        c_uint32,
        [c_int16, c_uint32, c_void_p, c_uint32, c_int32, c_uint32, c_void_p],
    )
    def_symbol(
        lib,
        "ps2000aGetValuesOverlappedBulk",
        c_uint32,
        [c_int16, c_uint32, c_void_p, c_uint32, c_int32, c_uint32, c_int32, c_void_p],
    )
    def_symbol(
        lib,
        "ps2000aStop",
        c_uint32,
        [
            c_int16,
        ],
    )
    def_symbol(lib, "ps2000aHoldOff", c_uint32, [c_int16, c_uint64, c_int32])
    def_symbol(
        lib,
        "ps2000aGetChannelInformation",
        c_uint32,
        [c_int16, c_int32, c_int32, c_int32, c_int32, c_int32],
    )
    def_symbol(lib, "ps2000aEnumerateUnits", c_uint32, [c_void_p, c_void_p, c_void_p])
    def_symbol(
        lib,
        "ps2000aPingUnit",
        c_uint32,
        [
            c_int16,
        ],
    )
    def_symbol(lib, "ps2000aMaximumValue", c_uint32, [c_int16, c_void_p])
    def_symbol(lib, "ps2000aMinimumValue", c_uint32, [c_int16, c_void_p])
    def_symbol(
        lib,
        "ps2000aGetAnalogueOffset",
        c_uint32,
        [c_int16, c_int32, c_int32, c_void_p, c_void_p],
    )
    def_symbol(lib, "ps2000aGetMaxSegments", c_uint32, [c_int16, c_void_p])

    return lib


_LIB_PS2000A = None


def load_ps2000a():
    global _LIB_PS2000A
    if _LIB_PS2000A is None:
        _LIB_PS2000A = _load_ps2000a_lib()
        return _LIB_PS2000A
    else:
        return _LIB_PS2000A


###
# Copyright CEA (Commissariat à l'énergie atomique et aux
# énergies alternatives) (2017-2025)
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
###
from __future__ import annotations

import abc
import ctypes
import logging
import time
from dataclasses import dataclass
from typing import Any, Optional, Sequence, Tuple

import numpy as np

from secbench.api.enums import Arithmetic, Coupling, Decimation, Slope
from secbench.api.exceptions import (
    InstrumentError,
    InstrumentUnsupportedFeature,
    InvalidParameter,
    NoSuchChannelError,
)
from secbench.api.instrument import Scope, ScopeAnalogChannel

from .error import PartialConfigurationError, PicoStatus, pico_check
from .types import ChannelState, TimebaseInfo, VerticalRange

logger = logging.getLogger(__name__)


@dataclass
class PicoHandles:
    psOpenUnit: Any
    psCloseUnit: Any
    psStop: Any
    psIsReady: Any
    psGetUnitInfo: Any
    psGetTimebase2: Any
    psSetChannel: Any
    psMemorySegments: Any
    psSetNoOfCaptures: Any
    psRunBlock: Any
    psSetDataBuffer: Any
    psSetDataBufferBulk: Any
    psGetValues: Any
    psGetValuesBulk: Any
    psSetSimpleTrigger: Any
    psPingUnit: Any = None


class PicobaseAnalogChannel(ScopeAnalogChannel):
    def __init__(self, parent: Picobase, name: str):
        self._parent = parent
        self._name = name
        self._state: ChannelState | None = None

    @property
    def parent(self) -> Scope:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    def state(self) -> ChannelState | None:
        return self._state

    def enabled(self) -> bool:
        if self._state is None:
            return False
        return self._state.enabled

    def coupling(self) -> Coupling:
        return self._state.coupling

    def range(self) -> float:
        return self._state.range

    def offset(self) -> float:
        return self._state.offset

    def decimation(self) -> Decimation:
        return Decimation.sample

    def disable(self) -> None:
        self._parent.pico_setup_channel(
            name=self._name,
            range=1,
            coupling=Coupling.dc,
            offset=0,
            enabled=False,
            decimation=Decimation.sample,
        )
        self._state = None

    def setup(
        self,
        range: float | None = None,
        coupling: Coupling | None = None,
        offset: float | None = None,
        decimation: Decimation = Decimation.sample,
    ) -> None:
        if self._state is not None:
            if range is None:
                range = self._state.range
            if coupling is None:
                coupling = self._state.coupling
            if offset is None:
                offset = self._state.offset
            # FIXME: we do not implement decimation modes currently.
            # if decimation is None:
            #     decimation = self._state.decimation
        else:
            if any(x is None for x in [range, coupling, offset, decimation]):
                raise ValueError(
                    f"Since channel {self._name} was disabled, all parameters (range, coupling, offset,"
                    f" decimation) must be explicitly given to the first setup call."
                )
        self._state = self._parent.pico_setup_channel(
            name=self._name,
            range=range,
            coupling=coupling,
            offset=offset,
            decimation=decimation,
            enabled=True,
        )

    def set_arithmetic(self, arithmetic: Arithmetic, reset: int = 1):
        raise NotImplementedError()


class Picobase(Scope, abc.ABC):
    """
    A base class that helps in implementing picoscope devices.

    .. note::

        Tips for implementors are prefixed with "for implementors" across the code.

    """

    # For implementors: overwrite this variable in subclasses with channel
    # name (str) -> API code
    _PICO_CHANNELS = {}

    # For implementors: overwrite this variable in subclasses with supported vertical
    # ranges.
    _PICO_VERTICAL_RANGES = []

    # Conversion from secbench's slope to API identifier.
    #
    # For implementors: check that the default implementation matches your device.
    _PICO_SLOPES = {}

    # For implementors: check the encoding are the same for your device.
    _PICO_COUPLINGS = {"DC50": 2, "DC": 1, "AC": 0, Coupling.dc: 1, Coupling.ac: 0}

    # For implementors: check the encoding are the same for your device.
    _PICO_BANDWIDTH = {
        "FULL": 0,
        "20MHZ": 1,
        "25MHZ": 2,
    }

    # For implementors: check the encoding are the same for your device.
    _PICO_RATIO_MODE = {
        "NONE": 0,
        "AGGREGATE": 1,
        "AVERAGE": 2,
        "DECIMATE": 4,
        "DISTRIBUTION": 8,
    }

    def __init__(
        self, lib, methods: PicoHandles, device_info_cls=None, serial_number=None
    ):
        self._handle = None
        handle = ctypes.c_int16(0)
        pico_check(methods.psOpenUnit(ctypes.pointer(handle), serial_number))
        assert handle.value > 0, "picoscope device handle is strictly positive"
        self._handle = handle
        self._lib = lib
        self._methods: PicoHandles = methods
        self._device_info_cls = device_info_cls
        self._device_info = {}

        # Check the device works.
        if self._methods.psPingUnit:
            pico_check(self._methods.psPingUnit(self._handle))

        if self._device_info_cls:
            for k in self._device_info_cls:
                assert isinstance(k, int), "device information are integers API code"
                value = self._pico_query_unit_info(k)
                if value is None:
                    logger.debug(
                        f"unable to query device information {k} (probably unsupported)"
                    )
                    continue
                self._device_info[k] = value

        # FIXME: duplicated code with segmented acquisition.
        max_samples = ctypes.c_uint32(0)
        pico_check(
            self._methods.psMemorySegments(self._handle, 1, ctypes.byref(max_samples))
        )
        self._max_samples = int(max_samples.value)
        self._acq_count = 1

        self._timebase: Optional[TimebaseInfo] = None
        self._samples = None
        adc_min, adc_max = self._pico_adc_range()
        self._adc_min = adc_min
        self._adc_max = adc_max

        self._buffers = {}

    def __del__(self):
        # The class may be partially initialized, we need to check if the load
        # was successful.
        self.close_unit()

    def close_unit(self):
        if self._handle is not None:
            logger.debug("closing unit")
            err = self._methods.psCloseUnit(self._handle)
            if err != 0:
                logger.warning("psCloseUnit failed with")
            self._handle = None

    def _pico_query_unit_info(self, key: int) -> Optional[str]:
        value_len = ctypes.c_int16(0)
        # Get the size of the info
        rc = self._methods.psGetUnitInfo(
            self._handle, None, 0, ctypes.byref(value_len), key
        )
        allowed = [
            PicoStatus.PICO_INFO_UNAVAILABLE.value,
            PicoStatus.PICO_INVALID_INFO.value,
        ]
        if rc in allowed:
            return None
        pico_check(rc)
        data = (ctypes.c_ubyte * value_len.value)()
        pico_check(
            self._methods.psGetUnitInfo(
                self._handle, data, len(data), ctypes.byref(value_len), key
            )
        )
        return bytes(data[:-1]).decode()

    def _pico_data_channels(self):
        return list(self._PICO_CHANNELS.keys())

    @abc.abstractmethod
    def _pico_adc_range(self) -> Tuple[int, int]:
        """
        Return the minimum and maximum supported ADC values.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def _pico_interval_to_timebase(x: float) -> int:
        """
        Convert duration to API timebase int.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def _pico_timebase_to_interval(x: int) -> float:
        """
        Convert API code to duration (in seconds).
        """
        pass

    def _pico_channel_to_api(self, ch_name: str) -> int:
        """
        Convert a channel name to its API identifier.

        The default implementation does a lookup in the :py:attr:`Picobase._PICO_CHANNELS`.
        """
        if ch_name not in self._PICO_CHANNELS:
            raise NoSuchChannelError(ch_name)
        return self._PICO_CHANNELS[ch_name]

    def _pico_supported_vertical_ranges(self) -> Sequence[VerticalRange]:
        """
        Get supported vertical ranges.

        The default implementation does a lookup in the :py:attr:`Picobase._PICO_VERTICAL_RANGES`.
        """
        return self._PICO_VERTICAL_RANGES

    def _pico_minimal_supported_range(self, requested_range) -> Optional[VerticalRange]:
        ranges = self._pico_supported_vertical_ranges()
        if requested_range > ranges[-1].volts:
            return None
        for v in self._pico_supported_vertical_ranges():
            if v.volts >= requested_range:
                return v
        return None

    def _pico_raw_to_volt(self, channel, raw):
        ch = self.channels()[channel]
        return ch.range() / self._adc_max * raw - ch.offset()

    def _pico_volt_to_raw(self, channel, volt):
        ch = self.channels()[channel]
        return int(self._adc_max * (volt + ch.offset()) / ch.range())

    def device_info(self):
        """
        Return all information available on the device.
        """
        return self._device_info

    def is_ready(self) -> bool:
        """
        Check if an acquisition is ready (i.e., data is available for retrival).
        """
        ready = ctypes.c_int16(0)
        pico_check(self._methods.psIsReady(self._handle, ctypes.byref(ready)))
        return ready.value != 0

    def stop_capture(self):
        """
        Stop any capture pending on the instrument.
        """
        pico_check(self._methods.psStop(self._handle))

    def set_sampling_interval(self, interval, duration, segment_index=0):
        # Find the closest matching parameters.
        timebase_id = self._pico_interval_to_timebase(interval)
        timebase_dt = self._pico_timebase_to_interval(timebase_id)
        requested_samples = int(round(duration / timebase_dt))

        interval_ns = ctypes.c_float(0.0)
        max_samples = ctypes.c_uint32(0)

        pico_check(
            self._methods.psGetTimebase2(
                self._handle,
                timebase_id,
                requested_samples,
                ctypes.byref(interval_ns),
                0,
                ctypes.byref(max_samples),
                segment_index,
            )
        )
        max_samples = max_samples.value
        if max_samples < requested_samples:
            logger.info(
                f"cannot acquire {max_samples} samples (too large, maximum value is {requested_samples})"
            )
            # FIXME(@TH): should we raise an error in that case?
        self._samples = requested_samples
        self._timebase = TimebaseInfo(
            timebase_id,
            float(interval_ns.value) * 1e-9,
            "",
            int(max_samples),
            segment_index,
        )

    def _check_horizontal_configuration(self):
        if self._timebase is None:
            raise PartialConfigurationError(
                "no horizontal configuration defined, please call horizontal() first"
            )

    # ===
    # Scope Interface
    # ===

    # Not implemented, should be handled by each device
    # def channels(self):
    #     ...

    def horizontal_interval(self):
        self._check_horizontal_configuration()
        return self._timebase.time_interval

    def horizontal_duration(self):
        self._check_horizontal_configuration()
        return self._samples * self._timebase.time_interval

    def horizontal_samples(self):
        self._check_horizontal_configuration()
        return self._samples

    def _alloc_memory_buffers(self):
        """
        This function is responsible for allocating and assigning memory buffers
        for segmented acquisition (aka. "rapid mode" in picoscope terminology).
        """
        if self._acq_count == 1:
            # Segmented acquisition is off, nothing to allocate.
            return

        samples = self.horizontal_samples()
        channels = [ch.name for ch in self.channels().values() if ch.enabled()]
        logger.debug(
            f"allocating memory buffers for segmented acquisition (channels, acq_count, samples) = ({len(channels)}, {self._acq_count}, {self._samples})"
        )
        for ch in channels:
            buffer = np.zeros((self._acq_count, samples), dtype=np.int16)
            assert buffer.data.c_contiguous
            ptr = buffer.ctypes.data
            self._buffers[ch] = buffer
            ch_number = self._PICO_CHANNELS[ch]
            for segment_index in range(self._acq_count):
                offset = ptr + 2 * segment_index * samples
                self._methods.psSetDataBufferBulk(
                    self._handle,
                    ch_number,
                    offset,
                    samples,
                    segment_index,
                    self._PICO_RATIO_MODE["NONE"],
                )

    def enable_segmented_acquisition(self, count: int):
        max_samples = ctypes.c_uint32(0)
        rc = self._methods.psMemorySegments(
            self._handle, count, ctypes.byref(max_samples)
        )
        pico_check(rc)
        self._max_samples = max_samples
        self._acq_count = count
        self._alloc_memory_buffers()

    def disable_segmented_acquisition(self):
        self._acq_count = 0

    def trigger_count(self) -> int:
        raise NotImplementedError()

    def disable_trigger_out(self):
        raise NotImplementedError()

    def reset(self):
        for ch in self.channels().values():
            ch.disable()
        self._timebase: Optional[TimebaseInfo] = None

    def sync(self):
        # Nothing to do.
        pass

    def _clear(self, pop_errors: bool):
        # Nothing to do.
        pass

    def bit_resolution(self) -> int:
        return 16

    def set_bit_resolution(self, prec: int):
        if prec != 16:
            logger.error(f"unsupported precision requested: {prec}")
            raise InvalidParameter("ps6000 only supports 16 bits precision")

    def pico_setup_channel(
        self,
        *,
        name: str,
        range: float,
        coupling: Coupling,
        offset: float,
        decimation: Decimation,
        enabled: bool = True,
    ) -> ChannelState:
        assert name in self.channels()
        vrange = self._pico_minimal_supported_range(range)
        if vrange is None:
            raise InvalidParameter(f"requested voltage range is too high: {range}")
        offset_api = ctypes.c_float(offset)
        logger.debug(
            f"channel {name}: requested {range} volts, using {vrange.user_str}"
        )
        rc = self._methods.psSetChannel(
            self._handle,
            self._PICO_CHANNELS[name],
            int(enabled),
            self._PICO_COUPLINGS[coupling],
            vrange.api_code,
            offset_api,
            self._PICO_BANDWIDTH["FULL"],
        )
        pico_check(rc)
        return ChannelState(
            enabled=enabled,
            range=vrange.volts,
            offset=offset,
            max_voltage=vrange.volts,
            coupling=coupling,
        )

    def _run_block(self, pre_samples=0, segment_index=0, oversample=1) -> float:
        self._check_horizontal_configuration()

        if self._acq_count > 1:
            rc = self._methods.psSetNoOfCaptures(
                self._handle, ctypes.c_uint32(self._acq_count)
            )
            pico_check(rc)
        assert pre_samples < self._samples
        post_samples = self._samples - pre_samples
        assert post_samples > 0
        hold_time_ms = ctypes.c_int32(0)
        rc = self._methods.psRunBlock(
            self._handle,
            pre_samples,
            post_samples,
            self._timebase.timebase_id,
            oversample,
            ctypes.byref(hold_time_ms),
            segment_index,
            None,
            None,
        )
        pico_check(rc)
        return hold_time_ms.value / 1e3

    def _arm(self, count: int, iterations: int, poll_interval: float) -> float:
        assert count == 1
        t = time.perf_counter()
        self._run_block()
        return time.perf_counter() - t

    def _wait(self, iterations: int, poll_interval: float) -> float:
        t = time.perf_counter()
        for _ in range(iterations):
            if self.is_ready():
                return time.perf_counter() - t
            time.sleep(poll_interval)
        logger.debug("wait timeout reached, stopping capture.")
        self.stop_capture()
        raise InstrumentError(
            "Timeout reached: instrument is still waiting for acquisitions."
        )

    def _wait_auto(self) -> float:
        raise NotImplementedError()

    def _set_data_format(self, bits: int, little_endian: bool):
        if bits != 16 and not little_endian:
            logger.warning(
                f"unsupported data format requested ({bits}, {little_endian}), ignored"
            )

    def _horizontal(self, interval=None, duration=None, samples=None):
        if duration and samples:
            interval = duration / samples
        elif interval and samples:
            duration = interval * samples
        self.set_sampling_interval(interval, duration)
        self._alloc_memory_buffers()

    def _get_data_segmented(self, channels):
        n_samples = ctypes.c_uint32(self._samples)
        overflow = (ctypes.c_int16 * self._acq_count)()
        pico_check(
            self._methods.psGetValuesBulk(
                self._handle,
                ctypes.byref(n_samples),  # noOfSamples
                0,  # fromSegmentIndex
                self._acq_count - 1,  # toSegmentIndex
                1,  # downSampleRatio
                self._PICO_RATIO_MODE["NONE"],  # downSampleRatioMode
                overflow,  # overflow
            )
        )
        return [self._buffers[ch] for ch in channels]

    def _get_data_single(self, channels, segment_index):
        data = [np.zeros(self._samples, dtype=np.int16) for _ in channels]

        for ch, d in zip(channels, data):
            ptr = d.ctypes.data
            ch_api = self._PICO_CHANNELS[ch]
            pico_check(
                self._methods.psSetDataBuffer(
                    self._handle, ch_api, ptr, len(d), self._PICO_RATIO_MODE["NONE"]
                )
            )
        n_samples = ctypes.c_uint32(self._samples)
        overflow = ctypes.c_int16(0)
        pico_check(
            self._methods.psGetValues(
                self._handle,
                0,
                ctypes.byref(n_samples),
                1,
                self._PICO_RATIO_MODE["NONE"],
                segment_index,
                ctypes.byref(overflow),
            )
        )
        n_samples = int(n_samples.value)
        if n_samples < self._samples:
            logger.info(
                f"got less samples than requested ({n_samples}, {self._samples} requested"
            )
        return data

    def _get_data(self, channels, volts: bool, segment_index=0):
        if self._acq_count > 1:
            data = self._get_data_segmented(channels)
        else:
            data = self._get_data_single(channels, segment_index=segment_index)

        for d in data:
            yield d

    def _setup_simple_trigger(
        self,
        channel,
        threshold=0,
        direction=Slope.rising,
        delay=0,
        timeout=0,
        enabled=True,
    ):
        ch_name = self._PICO_CHANNELS[channel]
        slope = self._PICO_SLOPES[direction]

        threshold_raw = self._pico_volt_to_raw(channel, threshold)
        delay = int(delay / self._timebase.time_interval)
        status = self._methods.psSetSimpleTrigger(
            self._handle,
            int(enabled),
            ch_name,
            threshold_raw,
            slope,
            delay,
            timeout,
        )
        pico_check(status)

    def _enable_trigger_out(self, slope, length, delay):
        raise InstrumentUnsupportedFeature("trigger out")

    def _set_trigger(self, channel, slope, level, delay):
        self._setup_simple_trigger(
            channel, threshold=level, direction=slope, delay=delay, timeout=0
        )


###
# Copyright CEA (Commissariat à l'énergie atomique et aux
# énergies alternatives) (2017-2025)
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
###

import ctypes
from enum import Enum

from .error import pico_check


def make_enum(name, fields):
    mappings = {}
    for i, f in enumerate(fields):
        mappings[i] = f
    return Enum(name, mappings)


def pico_enumerate_serial(ps_enumerate_handle):
    from .error import PicoscopeApiError

    try:
        count = ctypes.c_int16(0)
        serial_len = ctypes.c_int16(4096)
        serial_buff = (ctypes.c_int8 * serial_len.value)()
        pico_check(
            ps_enumerate_handle(
                ctypes.byref(count), serial_buff, ctypes.byref(serial_len)
            )
        )

        serial_ids = bytes(serial_buff[: serial_len.value - 1]).decode()
        for serial_num in serial_ids.split(","):
            yield (serial_num,)
    except PicoscopeApiError:
        pass


###
# Copyright CEA (Commissariat à l'énergie atomique et aux
# énergies alternatives) (2017-2025)
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
###

import pytest

from secbench.api import get_bench
from secbench.api.enums import Coupling, Decimation, Slope
from secbench.api.exceptions import InvalidParameter
from secbench.picoscope import PicoPS2000AScope, version
from secbench.picoscope.error import PicoscopeApiError


def test_version():
    assert version() == "0.2.1"


def has_scope() -> bool:
    hw = get_bench().get(PicoPS2000AScope, required=False)
    return hw is not None


pytestmark = pytest.mark.skipif(not has_scope(), reason="no PS2000AScope available")


@pytest.fixture
def scope():
    return get_bench().get(PicoPS2000AScope)


def test_description(scope):
    assert "picoscope" in scope.description.lower()


def test_channels(scope):
    assert set(scope.channels()) == {"A", "B"}


def test_reset_defaults(scope):
    scope.reset()
    for channel in scope.channels().values():
        assert not channel.enabled()


@pytest.mark.parametrize("channel", ["A", "B"])
def test_right_sequence_order(scope, channel):
    scope.reset()
    channel = scope[channel]
    channel.setup(range=1, offset=0, coupling=Coupling.dc, decimation=Decimation.sample)
    scope.set_horizontal(samples=1000, duration=1)


def test_enable_disable_channel(scope):
    scope.reset()
    for channel in scope.channels().values():
        assert not channel.enabled()
        channel.setup(range=1, coupling=Coupling.dc, offset=0)
        assert channel.enabled()
        channel.disable()
        assert not channel.enabled()


@pytest.mark.parametrize("range", [5, 1, 1e-1, 1e-2])
@pytest.mark.parametrize("offset", [-0.5, -0.1, 0, 0.1, 0.5])
@pytest.mark.parametrize("coupling", [Coupling.ac, Coupling.dc])
def test_channel_setup(scope, range, offset, coupling):
    scope.reset()
    for channel in scope.channels().values():
        channel.setup(range=range, offset=offset * range, coupling=coupling)
        assert channel.coupling() == coupling
        assert channel.range() >= range
        assert channel.offset() == pytest.approx(offset * range)
        assert channel.enabled()


def test_invalid_channel_setup(scope):
    scope.reset()
    channel = scope.channel_list()[0]

    with pytest.raises(InvalidParameter) as excinfo:
        channel.setup(
            range=200, offset=0, coupling=Coupling.dc, decimation=Decimation.sample
        )
    assert "voltage range is too high" in str(excinfo.value)

    with pytest.raises(PicoscopeApiError):
        channel.setup(range=1, coupling=Coupling.dc, offset=50)


@pytest.mark.parametrize("horizontal_args", [(1e-6, 1000, None), (1e-8, None, 1e-6)])
@pytest.mark.parametrize("channel", ["A", "B"])
def test_set_horizontal(scope, horizontal_args, channel):
    scope.reset()
    channel = scope[channel]
    interval, samples, duration = horizontal_args
    channel.setup(range=1, offset=0, coupling=Coupling.dc, decimation=Decimation.sample)
    scope.set_horizontal(interval=interval, samples=samples, duration=duration)
    if interval:
        assert scope.horizontal_interval() == pytest.approx(interval, abs=1e-2)
    if samples:
        assert scope.horizontal_samples() == pytest.approx(samples, abs=1e-2)
    if duration:
        assert scope.horizontal_duration() == pytest.approx(duration, abs=1e-2)


@pytest.mark.parametrize("channel", ["A", "B"])
@pytest.mark.parametrize("slope", [Slope.rising, Slope.falling, Slope.either])
@pytest.mark.parametrize("level", [-0.5, 0, 0.5, 1])
@pytest.mark.parametrize("delay", [0, 0.2])
def test_trigger(scope, slope, level, delay, channel):
    scope.reset()
    scope[channel].setup(range=10, coupling=Coupling.dc, offset=0)
    scope.set_horizontal(interval=8e-9, duration=80e-6)
    scope.set_trigger(channel, slope=slope, level=level, delay=delay)


###
# Copyright CEA (Commissariat à l'énergie atomique et aux
# énergies alternatives) (2017-2025)
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
###

from enum import Enum

from secbench.api.exceptions import InstrumentError

PICO_STATUS = {
    "PICO_OK": 0x00000000,
    "PICO_MAX_UNITS_OPENED": 0x00000001,
    "PICO_MEMORY_FAIL": 0x00000002,
    "PICO_NOT_FOUND": 0x00000003,
    "PICO_FW_FAIL": 0x00000004,
    "PICO_OPEN_OPERATION_IN_PROGRESS": 0x00000005,
    "PICO_OPERATION_FAILED": 0x00000006,
    "PICO_NOT_RESPONDING": 0x00000007,
    "PICO_CONFIG_FAIL": 0x00000008,
    "PICO_KERNEL_DRIVER_TOO_OLD": 0x00000009,
    "PICO_EEPROM_CORRUPT": 0x0000000A,
    "PICO_OS_NOT_SUPPORTED": 0x0000000B,
    "PICO_INVALID_HANDLE": 0x0000000C,
    "PICO_INVALID_PARAMETER": 0x0000000D,
    "PICO_INVALID_TIMEBASE": 0x0000000E,
    "PICO_INVALID_VOLTAGE_RANGE": 0x0000000F,
    "PICO_INVALID_CHANNEL": 0x00000010,
    "PICO_INVALID_TRIGGER_CHANNEL": 0x00000011,
    "PICO_INVALID_CONDITION_CHANNEL": 0x00000012,
    "PICO_NO_SIGNAL_GENERATOR": 0x00000013,
    "PICO_STREAMING_FAILED": 0x00000014,
    "PICO_BLOCK_MODE_FAILED": 0x00000015,
    "PICO_NULL_PARAMETER": 0x00000016,
    "PICO_ETS_MODE_SET": 0x00000017,
    "PICO_DATA_NOT_AVAILABLE": 0x00000018,
    "PICO_STRING_BUFFER_TO_SMALL": 0x00000019,
    "PICO_ETS_NOT_SUPPORTED": 0x0000001A,
    "PICO_AUTO_TRIGGER_TIME_TO_SHORT": 0x0000001B,
    "PICO_BUFFER_STALL": 0x0000001C,
    "PICO_TOO_MANY_SAMPLES": 0x0000001D,
    "PICO_TOO_MANY_SEGMENTS": 0x0000001E,
    "PICO_PULSE_WIDTH_QUALIFIER": 0x0000001F,
    "PICO_DELAY": 0x00000020,
    "PICO_SOURCE_DETAILS": 0x00000021,
    "PICO_CONDITIONS": 0x00000022,
    "PICO_USER_CALLBACK": 0x00000023,
    "PICO_DEVICE_SAMPLING": 0x00000024,
    "PICO_NO_SAMPLES_AVAILABLE": 0x00000025,
    "PICO_SEGMENT_OUT_OF_RANGE": 0x00000026,
    "PICO_BUSY": 0x00000027,
    "PICO_STARTINDEX_INVALID": 0x00000028,
    "PICO_INVALID_INFO": 0x00000029,
    "PICO_INFO_UNAVAILABLE": 0x0000002A,
    "PICO_INVALID_SAMPLE_INTERVAL": 0x0000002B,
    "PICO_TRIGGER_ERROR": 0x0000002C,
    "PICO_MEMORY": 0x0000002D,
    "PICO_SIG_GEN_PARAM": 0x0000002E,
    "PICO_SHOTS_SWEEPS_WARNING": 0x0000002F,
    "PICO_SIGGEN_TRIGGER_SOURCE": 0x00000030,
    "PICO_AUX_OUTPUT_CONFLICT": 0x00000031,
    "PICO_AUX_OUTPUT_ETS_CONFLICT": 0x00000032,
    "PICO_WARNING_EXT_THRESHOLD_CONFLICT": 0x00000033,
    "PICO_WARNING_AUX_OUTPUT_CONFLICT": 0x00000034,
    "PICO_SIGGEN_OUTPUT_OVER_VOLTAGE": 0x00000035,
    "PICO_DELAY_NULL": 0x00000036,
    "PICO_INVALID_BUFFER": 0x00000037,
    "PICO_SIGGEN_OFFSET_VOLTAGE": 0x00000038,
    "PICO_SIGGEN_PK_TO_PK": 0x00000039,
    "PICO_CANCELLED": 0x0000003A,
    "PICO_SEGMENT_NOT_USED": 0x0000003B,
    "PICO_INVALID_CALL": 0x0000003C,
    "PICO_GET_VALUES_INTERRUPTED": 0x0000003D,
    "PICO_NOT_USED": 0x0000003F,
    "PICO_INVALID_SAMPLERATIO": 0x00000040,
    "PICO_INVALID_STATE": 0x00000041,
    "PICO_NOT_ENOUGH_SEGMENTS": 0x00000042,
    "PICO_DRIVER_FUNCTION": 0x00000043,
    "PICO_RESERVED": 0x00000044,
    "PICO_INVALID_COUPLING": 0x00000045,
    "PICO_BUFFERS_NOT_SET": 0x00000046,
    "PICO_RATIO_MODE_NOT_SUPPORTED": 0x00000047,
    "PICO_RAPID_NOT_SUPPORT_AGGREGATION": 0x00000048,
    "PICO_INVALID_TRIGGER_PROPERTY": 0x00000049,
    "PICO_INTERFACE_NOT_CONNECTED": 0x0000004A,
    "PICO_RESISTANCE_AND_PROBE_NOT_ALLOWED": 0x0000004B,
    "PICO_POWER_FAILED": 0x0000004C,
    "PICO_SIGGEN_WAVEFORM_SETUP_FAILED": 0x0000004D,
    "PICO_FPGA_FAIL": 0x0000004E,
    "PICO_POWER_MANAGER": 0x0000004F,
    "PICO_INVALID_ANALOGUE_OFFSET": 0x00000050,
    "PICO_PLL_LOCK_FAILED": 0x00000051,
    "PICO_ANALOG_BOARD": 0x00000052,
    "PICO_CONFIG_FAIL_AWG": 0x00000053,
    "PICO_INITIALISE_FPGA": 0x00000054,
    "PICO_EXTERNAL_FREQUENCY_INVALID": 0x00000056,
    "PICO_CLOCK_CHANGE_ERROR": 0x00000057,
    "PICO_TRIGGER_AND_EXTERNAL_CLOCK_CLASH": 0x00000058,
    "PICO_PWQ_AND_EXTERNAL_CLOCK_CLASH": 0x00000059,
    "PICO_UNABLE_TO_OPEN_SCALING_FILE": 0x0000005A,
    "PICO_MEMORY_CLOCK_FREQUENCY": 0x0000005B,
    "PICO_I2C_NOT_RESPONDING": 0x0000005C,
    "PICO_NO_CAPTURES_AVAILABLE": 0x0000005D,
    "PICO_NOT_USED_IN_THIS_CAPTURE_MODE": 0x0000005E,
    "PICO_TOO_MANY_TRIGGER_CHANNELS_IN_USE": 0x0000005F,
    "PICO_INVALID_TRIGGER_DIRECTION": 0x00000060,
    "PICO_INVALID_TRIGGER_STATES": 0x00000061,
    "PICO_GET_DATA_ACTIVE": 0x00000103,
    "PICO_IP_NETWORKED": 0x00000104,
    "PICO_INVALID_IP_ADDRESS": 0x00000105,
    "PICO_IPSOCKET_FAILED": 0x00000106,
    "PICO_IPSOCKET_TIMEDOUT": 0x00000107,
    "PICO_SETTINGS_FAILED": 0x00000108,
    "PICO_NETWORK_FAILED": 0x00000109,
    "PICO_WS2_32_DLL_NOT_LOADED": 0x0000010A,
    "PICO_INVALID_IP_PORT": 0x0000010B,
    "PICO_COUPLING_NOT_SUPPORTED": 0x0000010C,
    "PICO_BANDWIDTH_NOT_SUPPORTED": 0x0000010D,
    "PICO_INVALID_BANDWIDTH": 0x0000010E,
    "PICO_AWG_NOT_SUPPORTED": 0x0000010F,
    "PICO_ETS_NOT_RUNNING": 0x00000110,
    "PICO_SIG_GEN_WHITENOISE_NOT_SUPPORTED": 0x00000111,
    "PICO_SIG_GEN_WAVETYPE_NOT_SUPPORTED": 0x00000112,
    "PICO_INVALID_DIGITAL_PORT": 0x00000113,
    "PICO_INVALID_DIGITAL_CHANNEL": 0x00000114,
    "PICO_INVALID_DIGITAL_TRIGGER_DIRECTION": 0x00000115,
    "PICO_SIG_GEN_PRBS_NOT_SUPPORTED": 0x00000116,
    "PICO_ETS_NOT_AVAILABLE_WITH_LOGIC_CHANNELS": 0x00000117,
    "PICO_WARNING_REPEAT_VALUE": 0x00000118,
    "PICO_POWER_SUPPLY_CONNECTED": 0x00000119,
    "PICO_POWER_SUPPLY_NOT_CONNECTED": 0x0000011A,
    "PICO_POWER_SUPPLY_REQUEST_INVALID": 0x0000011B,
    "PICO_POWER_SUPPLY_UNDERVOLTAGE": 0x0000011C,
    "PICO_CAPTURING_DATA": 0x0000011D,
    "PICO_USB3_0_DEVICE_NON_USB3_0_PORT": 0x0000011E,
    "PICO_NOT_SUPPORTED_BY_THIS_DEVICE": 0x0000011F,
    "PICO_INVALID_DEVICE_RESOLUTION": 0x00000120,
    "PICO_INVALID_NUMBER_CHANNELS_FOR_RESOLUTION": 0x00000121,
    "PICO_CHANNEL_DISABLED_DUE_TO_USB_POWERED": 0x00000122,
    "PICO_SIGGEN_DC_VOLTAGE_NOT_CONFIGURABLE": 0x00000123,
    "PICO_NO_TRIGGER_ENABLED_FOR_TRIGGER_IN_PRE_TRIG": 0x00000124,
    "PICO_TRIGGER_WITHIN_PRE_TRIG_NOT_ARMED": 0x00000125,
    "PICO_TRIGGER_WITHIN_PRE_NOT_ALLOWED_WITH_DELAY": 0x00000126,
    "PICO_TRIGGER_INDEX_UNAVAILABLE": 0x00000127,
    "PICO_AWG_CLOCK_FREQUENCY": 0x00000128,
    "PICO_TOO_MANY_CHANNELS_IN_USE": 0x00000129,
    "PICO_NULL_CONDITIONS": 0x0000012A,
    "PICO_DUPLICATE_CONDITION_SOURCE": 0x0000012B,
    "PICO_INVALID_CONDITION_INFO": 0x0000012C,
    "PICO_SETTINGS_READ_FAILED": 0x0000012D,
    "PICO_SETTINGS_WRITE_FAILED": 0x0000012E,
    "PICO_ARGUMENT_OUT_OF_RANGE": 0x0000012F,
    "PICO_HARDWARE_VERSION_NOT_SUPPORTED": 0x00000130,
    "PICO_DIGITAL_HARDWARE_VERSION_NOT_SUPPORTED": 0x00000131,
    "PICO_ANALOGUE_HARDWARE_VERSION_NOT_SUPPORTED": 0x00000132,
    "PICO_UNABLE_TO_CONVERT_TO_RESISTANCE": 0x00000133,
    "PICO_DUPLICATED_CHANNEL": 0x00000134,
    "PICO_INVALID_RESISTANCE_CONVERSION": 0x00000135,
    "PICO_INVALID_VALUE_IN_MAX_BUFFER": 0x00000136,
    "PICO_INVALID_VALUE_IN_MIN_BUFFER": 0x00000137,
    "PICO_SIGGEN_FREQUENCY_OUT_OF_RANGE": 0x00000138,
    "PICO_EEPROM2_CORRUPT": 0x00000139,
    "PICO_EEPROM2_FAIL": 0x0000013A,
    "PICO_SERIAL_BUFFER_TOO_SMALL": 0x0000013B,
    "PICO_SIGGEN_TRIGGER_AND_EXTERNAL_CLOCK_CLASH": 0x0000013C,
    "PICO_WARNING_SIGGEN_AUXIO_TRIGGER_DISABLED": 0x0000013D,
    "PICO_SIGGEN_GATING_AUXIO_NOT_AVAILABLE": 0x00000013E,
    "PICO_SIGGEN_GATING_AUXIO_ENABLED": 0x00000013F,
    "PICO_RESOURCE_ERROR": 0x00000140,
    "PICO_TEMPERATURE_TYPE_INVALID": 0x00000141,
    "PICO_TEMPERATURE_TYPE_NOT_SUPPORTED": 0x00000142,
    "PICO_TIMEOUT": 0x00000143,
    "PICO_DEVICE_NOT_FUNCTIONING": 0x00000144,
    "PICO_INTERNAL_ERROR": 0x00000145,
    "PICO_MULTIPLE_DEVICES_FOUND": 0x00000146,
    "PICO_WARNING_NUMBER_OF_SEGMENTS_REDUCED": 0x00000147,
    "PICO_CAL_PINS_STATES": 0x00000148,
    "PICO_CAL_PINS_FREQUENCY": 0x00000149,
    "PICO_CAL_PINS_AMPLITUDE": 0x0000014A,
    "PICO_CAL_PINS_WAVETYPE": 0x0000014B,
    "PICO_CAL_PINS_OFFSET": 0x0000014C,
    "PICO_PROBE_FAULT": 0x0000014D,
    "PICO_PROBE_IDENTITY_UNKNOWN": 0x0000014E,
    "PICO_PROBE_POWER_DC_POWER_SUPPLE_REQUIRED": 0x0000014F,
    "PICO_PROBE_NOT_POWERED_THROUGH_DC_POWER_SUPPLY": 0x00000150,
    "PICO_PROBE_CONFIG_FAILURE": 0x00000151,
    "PICO_PROBE_INTERACTION_CALLBACK": 0x00000152,
    "PICO_UNKNOWN_INTELLIGENT_PROBE": 0x00000153,
    "PICO_INTELLIGENT_PROBE_CORRUPT": 0x00000154,
    "PICO_PROBE_COLLECTION_NOT_STARTED": 0x00000155,
    "PICO_PROBE_POWER_CONSUMPTION_EXCEEDED": 0x00000156,
    "PICO_WARNING_PROBE_CHANNEL_OUT_OF_SYNC": 0x00000157,
    "PICO_ENDPOINT_MISSING": 0x00000158,
    "PICO_UNKNOWN_ENDPOINT_REQUEST": 0x00000159,
    "PICO_ADC_TYPE_ERROR": 0x0000015A,
    "PICO_FPGA2_FAILED": 0x0000015B,
    "PICO_FPGA2_DEVICE_STATUS": 0x0000015C,
    "PICO_ENABLED_PROGRAM_FPGA2_FAILED": 0x0000015D,
    "PICO_NO_CANNELS_OR_PORTS_ENABLED": 0x0000015E,
    "PICO_INVALID_RATIO_MODE": 0x0000015F,
    "PICO_READS_NOT_SUPPORTED_IN_CURRENT_CAPTURE_MODE": 0x00000160,
    "PICO_TRIGGER_READ_SELECTION_CHECK_FAILED": 0x00000161,
    "PICO_DATA_READ1_SELECTION_CHECK_FAILED": 0x00000162,
    "PICO_DATA_READ2_SELECTION_CHECK_FAILED": 0x00000164,
    "PICO_DATA_READ3_SELECTION_CHECK_FAILED": 0x00000168,
    "PICO_READ_SELECTION_OUT_OF_RANGE": 0x00000170,
    "PICO_MULTIPLE_RATIO_MODES": 0x00000171,
    "PICO_NO_SAMPLES_READ": 0x00000172,
    "PICO_RATIO_MODE_NOT_REQUESTED": 0x00000173,
    "PICO_NO_USER_READ_REQUESTS": 0x00000174,
    "PICO_ZERO_SAMPLES_INVALID": 0x00000175,
    "PICO_ANALOGUE_HARDWARE_MISSING": 0x00000176,
    "PICO_ANALOGUE_HARDWARE_PINS": 0x00000177,
    "PICO_ANALOGUE_SMPS_FAULT": 0x00000178,
    "PICO_DIGITAL_ANALOGUE_HARDWARE_CONFLICT": 0x00000179,
    "PICO_RATIO_MODE_BUFFER_NOT_SET": 0x0000017A,
    "PICO_RESOLUTION_NOT_SUPPORTED_BY_VARIENT": 0x0000017B,
    "PICO_THRESHOLD_OUT_OF_RANGE": 0x0000017C,
    "PICO_INVALID_SIMPLE_TRIGGER_DIRECTION": 0x0000017D,
    "PICO_AUX_NOT_SUPPORTED": 0x0000017E,
    "PICO_NULL_DIRECTIONS": 0x0000017F,
    "PICO_NULL_CHANNEL_PROPERTIES": 0x00000180,
    "PICO_TRIGGER_CHANNEL_NOT_ENABLED": 0x00000181,
    "PICO_CONDITION_HAS_NO_TRIGGER_PROPERTY": 0x00000182,
    "PICO_RATIO_MODE_TRIGGER_MASKING_INVALID": 0x00000183,
    "PICO_TRIGGER_DATA_REQUIRES_MIN_BUFFER_SIZE_OF_40_SAMPLES": 0x00000184,
    "PICO_NO_OF_CAPTURES_OUT_OF_RANGE": 0x00000185,
    "PICO_RATIO_MODE_SEGMENT_HEADER_DOES_NOT_REQUIRE_BUFFERS": 0x00000186,
    "PICO_FOR_SEGMENT_HEADER_USE_GETTRIGGERINFO": 0x00000187,
    "PICO_READ_NOT_SET": 0x00000188,
    "PICO_ADC_SETTING_MISMATCH": 0x00000189,
    "PICO_DATATYPE_INVALID": 0x0000018A,
    "PICO_RATIO_MODE_DOES_NOT_SUPPORT_DATATYPE": 0x0000018B,
    "PICO_CHANNEL_COMBINATION_NOT_VALID_IN_THIS_RESOLUTION": 0x0000018C,
    "PICO_USE_8BIT_RESOLUTION": 0x0000018D,
    "PICO_AGGREGATE_BUFFERS_SAME_POINTER": 0x0000018E,
    "PICO_OVERLAPPED_READ_VALUES_OUT_OF_RANGE": 0x0000018F,
    "PICO_OVERLAPPED_READ_SEGMENTS_OUT_OF_RANGE": 0x00000190,
    "PICO_CHANNELFLAGSCOMBINATIONS_ARRAY_SIZE_TOO_SMALL": 0x00000191,
    "PICO_CAPTURES_EXCEEDS_NO_OF_SUPPORTED_SEGMENTS": 0x00000192,
    "PICO_TIME_UNITS_OUT_OF_RANGE": 0x00000193,
    "PICO_NO_SAMPLES_REQUESTED": 0x00000194,
    "PICO_INVALID_ACTION": 0x00000195,
    "PICO_NO_OF_SAMPLES_NEED_TO_BE_EQUAL_WHEN_ADDING_BUFFERS": 0x00000196,
    "PICO_WAITING_FOR_DATA_BUFFERS": 0x00000197,
    "PICO_STREAMING_ONLY_SUPPORTS_ONE_READ": 0x00000198,
    "PICO_CLEAR_DATA_BUFFER_INVALID": 0x00000199,
    "PICO_INVALID_ACTION_FLAGS_COMBINATION": 0x0000019A,
    "PICO_PICO_MOTH_MIN_AND_MAX_NULL_BUFFERS_CANNOT_BE_ADDED": 0x0000019B,
    "PICO_CONFLICT_IN_SET_DATA_BUFFERS_CALL_REMOVE_DATA_BUFFER_TO_RESET": 0x0000019C,
    "PICO_REMOVING_DATA_BUFFER_ENTRIES_NOT_ALLOWED_WHILE_DATA_PROCESSING": 0x0000019D,
    "PICO_CYUSB_REQUEST_FAILED": 0x00000200,
    "PICO_STREAMING_DATA_REQUIRED": 0x00000201,
    "PICO_INVALID_NUMBER_OF_SAMPLES": 0x00000202,
    "PICO_INALID_DISTRIBUTION": 0x00000203,
    "PICO_BUFFER_LENGTH_GREATER_THAN_INT32_T": 0x00000204,
    "PICO_PLL_MUX_OUT_FAILED": 0x00000209,
    "PICO_ONE_PULSE_WIDTH_DIRECTION_ALLOWED": 0x0000020A,
    "PICO_EXTERNAL_TRIGGER_NOT_SUPPORTED": 0x0000020B,
    "PICO_NO_TRIGGER_CONDITIONS_SET": 0x0000020C,
    "PICO_NO_OF_CHANNEL_TRIGGER_PROPERTIES_OUT_OF_RANGE": 0x0000020D,
    "PICO_PROBE_COMPNENT_ERROR": 0x0000020E,
    "PICO_INVALID_TRIGGER_CHANNELS_FOR_ETS": 0x00000210,
    "PICO_NOT_AVALIABLE_WHEN_STREAMING_IS_RUNNING": 0x00000211,
    "PICO_INVALID_TRIGGER_WITHIN_PRE_TRIGGER_STATE": 0x00000212,
    "PICO_ZERO_NUMBER_OF_CAPTURES_INVALID": 0x00000213,
    "PICO_TRIGGER_DELAY_OUT_OF_RANGE": 0x00000300,
    "PICO_INVALID_THRESHOLD_DIRECTION": 0x00000301,
    "PICO_INVALID_THRESGOLD_MODE": 0x00000302,
    "PICO_DEVICE_TIME_STAMP_RESET": 0x01000000,
    "PICO_WATCHDOGTIMER": 0x10000000,
    "PICO_IPP_NOT_FOUND": 0x10000001,
    "PICO_IPP_NO_FUNCTION": 0x10000002,
    "PICO_IPP_ERROR": 0x10000003,
    "PICO_SHADOW_CAL_NOT_AVAILABLE": 0x10000004,
    "PICO_SHADOW_CAL_DISABLED": 0x10000005,
    "PICO_SHADOW_CAL_ERROR": 0x10000006,
    "PICO_SHADOW_CAL_CORRUPT": 0x10000007,
}
PICO_STATUS_STR = dict((v, k) for k, v in PICO_STATUS.items())

PicoStatus = Enum("PicoStatus", PICO_STATUS)


class PicoscopeApiError(InstrumentError):
    """
    An error in the picoscope API.
    """

    def __init__(self, code, context=""):
        super().__init__(f"Error code returned by the picoscope device (code={code})")
        self.code = code
        self.context = context

    def __str__(self):
        return repr(self)

    def __repr__(self):
        err_str = PICO_STATUS_STR.get(self.code, "???")
        return f"<PicoscopeApiError: {self.code} {err_str})>"


def pico_check(code, context=""):
    if code != 0:
        raise PicoscopeApiError(code, context=context)


class PartialConfigurationError(InstrumentError):
    def __init__(self, message):
        super().__init__(message)


###
# Copyright CEA (Commissariat à l'énergie atomique et aux
# énergies alternatives) (2017-2025)
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
###

"""
PS6000 driver only using direct library calls
"""

from ctypes import (
    c_char_p,
    c_double,
    c_float,
    c_int16,
    c_int32,
    c_uint16,
    c_uint32,
    c_void_p,
)
from enum import IntEnum

from .common import def_symbol, load_libps


class PicoInfo(IntEnum):
    """
    Information to be requested to `ps6000GetUnitInfo`.
    """

    DRIVER_VERSION = 0x00000000
    USB_VERSION = 0x00000001
    HARDWARE_VERSION = 0x00000002
    VARIANT_INFO = 0x00000003
    BATCH_AND_SERIAL = 0x00000004
    CAL_DATE = 0x00000005
    KERNEL_VERSION = 0x00000006
    DIGITAL_HARDWARE_VERSION = 0x00000007
    ANALOGUE_HARDWARE_VERSION = 0x00000008
    FIRMWARE_VERSION_1 = 0x00000009
    FIRMWARE_VERSION_2 = 0x0000000A
    MAC_ADDRESS = 0x0000000B
    SHADOW_CAL = 0x0000000C
    IPP_VERSION = 0x0000000D


def _load_ps6000_lib(name="ps6000"):
    lib = load_libps(name)

    # fmt: off
    doc = """ PICO_STATUS ps6000OpenUnit
        (
            int16_t *handle,
            int8_t  *serial
        ); """
    def_symbol(lib, "ps6000OpenUnit", c_uint32, [c_void_p, c_char_p], doc)

    doc = """ PICO_STATUS ps6000OpenUnitAsync
        (
            int16_t *status,
            int8_t  *serial
        ); """
    def_symbol(lib, "ps6000OpenUnitAsync", c_uint32, [c_void_p, c_char_p], doc)

    doc = """ PICO_STATUS ps6000OpenUnitProgress
        (
          int16_t *handle,
          int16_t *progressPercent,
          int16_t *complete
        ); """
    def_symbol(lib, "ps6000OpenUnitProgress", c_uint32, [c_void_p, c_void_p, c_void_p],
               doc)

    doc = """ PICO_STATUS ps6000GetUnitInfo
        (
            int16_t    handle,
            int8_t    *string,
            int16_t    stringLength,
            int16_t   *requiredSize,
            PICO_INFO  info
        ); """
    def_symbol(lib, "ps6000GetUnitInfo", c_uint32,
               [c_int16, c_void_p, c_int16, c_void_p, c_uint32], doc)

    doc = """ PICO_STATUS ps6000FlashLed
        (
            int16_t  handle,
            int16_t  start
        ); """
    def_symbol(lib, "ps6000FlashLed", c_uint32, [c_int16, c_int16], doc)

    doc = """ PICO_STATUS ps6000CloseUnit
        (
            int16_t  handle
        ); """
    def_symbol(lib, "ps6000CloseUnit", c_uint32, [c_int16, ], doc)

    doc = """ PICO_STATUS ps6000MemorySegments
        (
            int16_t   handle,
            uint32_t  nSegments,
            uint32_t *nMaxSamples
        ); """
    def_symbol(lib, "ps6000MemorySegments", c_uint32, [c_int16, c_uint32, c_void_p],
               doc)

    doc = """ PICO_STATUS ps6000SetChannel
        (
            int16_t                   handle,
            PS6000_CHANNEL            channel,
            int16_t                   enabled,
            PS6000_COUPLING           type,
            PS6000_RANGE              range,
            float                     analogueOffset,
            PS6000_BANDWIDTH_LIMITER  bandwidth
        ); """
    def_symbol(lib, "ps6000SetChannel", c_uint32,
               [c_int16, c_int32, c_int16, c_int32, c_int32, c_float, c_int32], doc)

    doc = """ PICO_STATUS ps6000GetTimebase
        (
            int16_t   handle,
            uint32_t  timebase,
            uint32_t  noSamples,
            int32_t  *timeIntervalNanoseconds,
            int16_t   oversample,
            uint32_t *maxSamples,
            uint32_t  segmentIndex
        ); """
    def_symbol(lib, "ps6000GetTimebase", c_uint32,
               [c_int16, c_uint32, c_uint32, c_void_p, c_int16, c_void_p, c_uint32],
               doc)

    doc = """ PICO_STATUS ps6000GetTimebase2
        (
            int16_t   handle,
            uint32_t  timebase,
            uint32_t  noSamples,
            float    *timeIntervalNanoseconds,
            int16_t   oversample,
            uint32_t *maxSamples,
            uint32_t  segmentIndex
        ); """
    def_symbol(lib, "ps6000GetTimebase2", c_uint32,
               [c_int16, c_uint32, c_uint32, c_void_p, c_int16, c_void_p, c_uint32],
               doc)

    doc = """ PICO_STATUS ps6000SetSigGenArbitrary
        (
            int16_t                    handle,
            int32_t                    offsetVoltage,
            uint32_t                   pkToPk,
            uint32_t                   startDeltaPhase,
            uint32_t                   stopDeltaPhase,
            uint32_t                   deltaPhaseIncrement,
            uint32_t                   dwellCount,
            int16_t                   *arbitraryWaveform,
            int32_t                    arbitraryWaveformSize,
            PS6000_SWEEP_TYPE          sweepType,
            PS6000_EXTRA_OPERATIONS    operation,
            PS6000_INDEX_MODE          indexMode,
            uint32_t                   shots,
            uint32_t                   sweeps,
            PS6000_SIGGEN_TRIG_TYPE    triggerType,
            PS6000_SIGGEN_TRIG_SOURCE  triggerSource,
            int16_t                    extInThreshold
        ); """
    def_symbol(lib, "ps6000SetSigGenArbitrary", c_uint32,
               [c_int16, c_int32, c_uint32, c_uint32, c_uint32, c_uint32, c_uint32,
                c_void_p, c_int32, c_int32, c_int32, c_int32, c_uint32, c_uint32,
                c_int32, c_int32, c_int16], doc)

    doc = """ PICO_STATUS ps6000SetSigGenBuiltIn
        (
            int16_t                    handle,
            int32_t                    offsetVoltage,
            uint32_t                   pkToPk,
            int16_t                    waveType,
            float                      startFrequency,
            float                      stopFrequency,
            float                      increment,
            float                      dwellTime,
            PS6000_SWEEP_TYPE          sweepType,
            PS6000_EXTRA_OPERATIONS    operation,
            uint32_t                   shots,
            uint32_t                   sweeps,
            PS6000_SIGGEN_TRIG_TYPE    triggerType,
            PS6000_SIGGEN_TRIG_SOURCE  triggerSource,
            int16_t                    extInThreshold
        ); """
    def_symbol(lib, "ps6000SetSigGenBuiltIn", c_uint32,
               [c_int16, c_int32, c_uint32, c_int16, c_float, c_float, c_float, c_float,
                c_int32, c_int32, c_uint32, c_uint32, c_int32, c_int32, c_int16], doc)

    doc = """ PICO_STATUS ps6000SetSigGenBuiltInV2
        (
            int16_t                    handle,
            int32_t                    offsetVoltage,
            uint32_t                   pkToPk,
            int16_t                    waveType,
            double                     startFrequency,
            double                     stopFrequency,
            double                     increment,
            double                     dwellTime,
            PS6000_SWEEP_TYPE          sweepType,
            PS6000_EXTRA_OPERATIONS    operation,
            uint32_t                   shots,
            uint32_t                   sweeps,
            PS6000_SIGGEN_TRIG_TYPE    triggerType,
            PS6000_SIGGEN_TRIG_SOURCE  triggerSource,
            int16_t                    extInThreshold
        ); """
    def_symbol(lib, "ps6000SetSigGenBuiltInV2", c_uint32,
               [c_int16, c_int32, c_uint32, c_int16, c_double, c_double, c_double,
                c_double, c_int32, c_int32, c_uint32, c_uint32, c_int32, c_int32,
                c_int16], doc)

    doc = """ PICO_STATUS ps6000SetSigGenPropertiesArbitrary
        (
            int16_t                    handle,
            int32_t                    offsetVoltage,
            uint32_t                   pkToPk,
            uint32_t                   startDeltaPhase,
            uint32_t                   stopDeltaPhase,
            uint32_t                   deltaPhaseIncrement,
            uint32_t                   dwellCount,
            PS6000_SWEEP_TYPE          sweepType,
            uint32_t                   shots,
            uint32_t                   sweeps,
            PS6000_SIGGEN_TRIG_TYPE    triggerType,
            PS6000_SIGGEN_TRIG_SOURCE  triggerSource,
            int16_t                    extInThreshold
        ); """
    def_symbol(lib, "ps6000SetSigGenPropertiesArbitrary", c_uint32,
               [c_int16, c_int32, c_uint32, c_uint32, c_uint32, c_uint32, c_uint32,
                c_int32, c_uint32, c_uint32, c_int32, c_int32, c_int16], doc)

    doc = """ PICO_STATUS ps6000SetSigGenPropertiesBuiltIn
        (
            int16_t                    handle,
            int32_t                    offsetVoltage,
            uint32_t                   pkToPk,
            double                     startFrequency,
            double                     stopFrequency,
            double                     increment,
            double                     dwellTime,
            PS6000_SWEEP_TYPE          sweepType,
            uint32_t                   shots,
            uint32_t                   sweeps,
            PS6000_SIGGEN_TRIG_TYPE    triggerType,
            PS6000_SIGGEN_TRIG_SOURCE  triggerSource,
            int16_t                    extInThreshold
        ); """
    def_symbol(lib, "ps6000SetSigGenPropertiesBuiltIn", c_uint32,
               [c_int16, c_int32, c_uint32, c_double, c_double, c_double, c_double,
                c_int32, c_uint32, c_uint32, c_int32, c_int32, c_int16], doc)

    doc = """ PICO_STATUS ps6000SigGenFrequencyToPhase
        (
            int16_t            handle,
            double             frequency,
            PS6000_INDEX_MODE  indexMode,
            uint32_t           bufferLength,
            uint32_t          *phase
        ); """
    def_symbol(lib, "ps6000SigGenFrequencyToPhase", c_uint32,
               [c_int16, c_double, c_int32, c_uint32, c_void_p], doc)

    doc = """ PICO_STATUS ps6000SigGenArbitraryMinMaxValues
        (
            int16_t   handle,
            int16_t  *minArbitraryWaveformValue,
            int16_t  *maxArbitraryWaveformValue,
            uint32_t *minArbitraryWaveformSize,
            uint32_t *maxArbitraryWaveformSize
        ); """
    def_symbol(lib, "ps6000SigGenArbitraryMinMaxValues", c_uint32,
               [c_int16, c_void_p, c_void_p, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps6000SigGenSoftwareControl
        (
            int16_t  handle,
            int16_t  state
        ); """
    def_symbol(lib, "ps6000SigGenSoftwareControl", c_uint32, [c_int16, c_int16], doc)

    doc = """ PICO_STATUS ps6000SetSimpleTrigger
        (
            int16_t                     handle,
            int16_t                     enable,
            PS6000_CHANNEL              source,
            int16_t                     threshold,
            PS6000_THRESHOLD_DIRECTION  direction,
            uint32_t                    delay,
            int16_t                     autoTrigger_ms
        ); """
    def_symbol(lib, "ps6000SetSimpleTrigger", c_uint32,
               [c_int16, c_int16, c_int32, c_int16, c_int32, c_uint32, c_int16], doc)

    doc = """ PICO_STATUS ps6000SetEts
        (
            int16_t          handle,
            PS6000_ETS_MODE  mode,
            int16_t          etsCycles,
            int16_t          etsInterleave,
            int32_t         *sampleTimePicoseconds
        ); """
    def_symbol(lib, "ps6000SetEts", c_uint32,
               [c_int16, c_int32, c_int16, c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps6000SetTriggerChannelProperties
        (
            int16_t                            handle,
            PS6000_TRIGGER_CHANNEL_PROPERTIES *channelProperties,
            int16_t                            nChannelProperties,
            int16_t                            auxOutputEnable,
            int32_t                            autoTriggerMilliseconds
        ); """
    def_symbol(lib, "ps6000SetTriggerChannelProperties", c_uint32,
               [c_int16, c_void_p, c_int16, c_int16, c_int32], doc)

    doc = """ PICO_STATUS ps6000SetTriggerChannelConditions
        (
            int16_t                    handle,
            PS6000_TRIGGER_CONDITIONS *conditions,
            int16_t                    nConditions
        ); """
    def_symbol(lib, "ps6000SetTriggerChannelConditions", c_uint32,
               [c_int16, c_void_p, c_int16], doc)

    doc = """ PICO_STATUS ps6000SetTriggerChannelDirections
        (
            int16_t                       handle,
            PS6000_THRESHOLD_DIRECTION  channelA,
            PS6000_THRESHOLD_DIRECTION  channelB,
            PS6000_THRESHOLD_DIRECTION  channelC,
            PS6000_THRESHOLD_DIRECTION  channelD,
            PS6000_THRESHOLD_DIRECTION  ext,
            PS6000_THRESHOLD_DIRECTION  aux
        ); """
    def_symbol(lib, "ps6000SetTriggerChannelDirections", c_uint32,
               [c_int16, c_int32, c_int32, c_int32, c_int32, c_int32, c_int32], doc)

    doc = """ PICO_STATUS ps6000SetTriggerDelay
        (
            int16_t   handle,
            uint32_t  delay
        ); """
    def_symbol(lib, "ps6000SetTriggerDelay", c_uint32, [c_int16, c_uint32], doc)

    doc = """ PICO_STATUS ps6000SetPulseWidthQualifier
        (
            int16_t                     handle,
            PS6000_PWQ_CONDITIONS      *conditions,
            int16_t                     nConditions,
            PS6000_THRESHOLD_DIRECTION  direction,
            uint32_t                    lower,
            uint32_t                    upper,
            PS6000_PULSE_WIDTH_TYPE     type
        ); """
    def_symbol(lib, "ps6000SetPulseWidthQualifier", c_uint32,
               [c_int16, c_void_p, c_int16, c_int32, c_uint32, c_uint32, c_int32], doc)

    doc = """ PICO_STATUS ps6000IsTriggerOrPulseWidthQualifierEnabled
        (
            int16_t  handle,
            int16_t *triggerEnabled,
            int16_t *pulseWidthQualifierEnabled
        ); """
    def_symbol(lib, "ps6000IsTriggerOrPulseWidthQualifierEnabled", c_uint32,
               [c_int16, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps6000GetTriggerTimeOffset
        (
            int16_t            handle,
            uint32_t          *timeUpper,
            uint32_t          *timeLower,
            PS6000_TIME_UNITS *timeUnits,
            uint32_t           segmentIndex
        ); """
    def_symbol(lib, "ps6000GetTriggerTimeOffset", c_uint32,
               [c_int16, c_void_p, c_void_p, c_void_p, c_uint32], doc)

    doc = """ PICO_STATUS ps6000GetTriggerTimeOffset64
        (
            int16_t              handle,
            int64_t           *time,
            PS6000_TIME_UNITS *timeUnits,
            uint32_t      segmentIndex
        ); """
    def_symbol(lib, "ps6000GetTriggerTimeOffset64", c_uint32,
               [c_int16, c_void_p, c_void_p, c_uint32], doc)

    doc = """ PICO_STATUS ps6000GetValuesTriggerTimeOffsetBulk
        (
            int16_t            handle,
            uint32_t          *timesUpper,
            uint32_t          *timesLower,
            PS6000_TIME_UNITS *timeUnits,
            uint32_t           fromSegmentIndex,
            uint32_t           toSegmentIndex
        ); """
    def_symbol(lib, "ps6000GetValuesTriggerTimeOffsetBulk", c_uint32,
               [c_int16, c_void_p, c_void_p, c_void_p, c_uint32, c_uint32], doc)

    doc = """ PICO_STATUS ps6000GetValuesTriggerTimeOffsetBulk64
        (
            int16_t            handle,
            int64_t           *times,
            PS6000_TIME_UNITS *timeUnits,
            uint32_t           fromSegmentIndex,
            uint32_t           toSegmentIndex
        ); """
    def_symbol(lib, "ps6000GetValuesTriggerTimeOffsetBulk64", c_uint32,
               [c_int16, c_void_p, c_void_p, c_uint32, c_uint32], doc)

    doc = """ PICO_STATUS ps6000SetDataBuffers
        (
            int16_t            handle,
            PS6000_CHANNEL     channel,
            int16_t           *bufferMax,
            int16_t           *bufferMin,
            uint32_t           bufferLth,
            PS6000_RATIO_MODE  downSampleRatioMode
        ); """
    def_symbol(lib, "ps6000SetDataBuffers", c_uint32,
               [c_int16, c_int32, c_void_p, c_void_p, c_uint32, c_int32], doc)

    doc = """ PICO_STATUS ps6000SetDataBuffer
        (
            int16_t            handle,
            PS6000_CHANNEL     channel,
            int16_t           *buffer,
            uint32_t           bufferLth,
            PS6000_RATIO_MODE  downSampleRatioMode
        ); """
    def_symbol(lib, "ps6000SetDataBuffer", c_uint32,
               [c_int16, c_int32, c_void_p, c_uint32, c_int32], doc)

    doc = """ PICO_STATUS (ps6000SetDataBufferBulk)
        (
            int16_t            handle,
            PS6000_CHANNEL     channel,
            int16_t           *buffer,
            uint32_t           bufferLth,
            uint32_t           waveform,
            PS6000_RATIO_MODE  downSampleRatioMode
        ); """
    def_symbol(lib, "ps6000SetDataBufferBulk", c_uint32,
               [c_int16, c_int32, c_void_p, c_uint32, c_uint32, c_int32], doc)

    doc = """ PICO_STATUS ps6000SetDataBuffersBulk
        (
            int16_t            handle,
            PS6000_CHANNEL     channel,
            int16_t           *bufferMax,
            int16_t           *bufferMin,
            uint32_t           bufferLth,
            uint32_t           waveform,
            PS6000_RATIO_MODE  downSampleRatioMode
        ); """
    def_symbol(lib, "ps6000SetDataBuffersBulk", c_uint32,
               [c_int16, c_int32, c_void_p, c_void_p, c_uint32, c_uint32, c_int32], doc)

    doc = """ PICO_STATUS ps6000SetEtsTimeBuffer
        (
            int16_t   handle,
            int64_t  *buffer,
            uint32_t  bufferLth
        ); """
    def_symbol(lib, "ps6000SetEtsTimeBuffer", c_uint32, [c_int16, c_void_p, c_uint32],
               doc)

    doc = """ PICO_STATUS ps6000SetEtsTimeBuffers
        (
            int16_t   handle,
            uint32_t *timeUpper,
            uint32_t *timeLower,
            uint32_t  bufferLth
        ); """
    def_symbol(lib, "ps6000SetEtsTimeBuffers", c_uint32,
               [c_int16, c_void_p, c_void_p, c_uint32], doc)

    doc = """ PICO_STATUS ps6000RunBlock
        (
            int16_t           handle,
            uint32_t          noOfPreTriggerSamples,
            uint32_t          noOfPostTriggerSamples,
            uint32_t          timebase,
            int16_t           oversample,
            int32_t          *timeIndisposedMs,
            uint32_t          segmentIndex,
            ps6000BlockReady  lpReady,
            void             *pParameter
        ); """
    def_symbol(lib, "ps6000RunBlock", c_uint32,
               [c_int16, c_uint32, c_uint32, c_uint32, c_int16, c_void_p, c_uint32,
                c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps6000IsReady
        (
            int16_t  handle,
            int16_t *ready
        ); """
    def_symbol(lib, "ps6000IsReady", c_uint32, [c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps6000RunStreaming
        (
            int16_t            handle,
            uint32_t          *sampleInterval,
            PS6000_TIME_UNITS  sampleIntervalTimeUnits,
            uint32_t           maxPreTriggerSamples,
            uint32_t           maxPostPreTriggerSamples,
            int16_t            autoStop,
            uint32_t           downSampleRatio,
            PS6000_RATIO_MODE  downSampleRatioMode,
            uint32_t           overviewBufferSize
        ); """
    def_symbol(lib, "ps6000RunStreaming", c_uint32,
               [c_int16, c_void_p, c_int32, c_uint32, c_uint32, c_int16, c_uint32,
                c_int32, c_uint32], doc)

    # FIXME: callbacks not implemented
    # doc = """ void ps6000BlockReady
    #     (
    #         int16_t          handle,
    #         PICO_STATUS      status,
    #         void             *pParameter
    #     ); """
    # ps6000.BlockReadyType = C_CALLBACK_FUNCTION_FACTORY(None,
    #                                                     c_int16,
    #                                                     c_uint32,
    #                                                     c_void_p)
    # ps6000.BlockReadyType.__doc__ = doc

    # FIXME: callbacks not implemented
    # doc = """ void ps6000StreamingReady
    #     (
    #             int16_t                         handle,
    #             uint32_t                        noOfSamples,
    #             uint32_t                        startIndex,
    #             int16_t                         overflow,
    #             uint32_t                        triggerAt,
    #             int16_t                         triggered,
    #             int16_t                         autoStop,
    #             void                            *pParameter
    #     ); """
    # ps6000.StreamingReadyType = C_CALLBACK_FUNCTION_FACTORY(None,
    #                                                         c_int16,
    #                                                         c_uint32,
    #                                                         c_uint32,
    #                                                         c_int16,
    #                                                         c_uint32,
    #                                                         c_int16,
    #                                                         c_int16,
    #                                                         c_void_p)

    # ps6000.StreamingReadyType.__doc__ = doc

    doc = """ PICO_STATUS ps6000GetStreamingLatestValues
        (
            int16_t               handle,
            ps6000StreamingReady  lpPs6000Ready,
            void                 *pParameter
        ); """
    def_symbol(lib, "ps6000GetStreamingLatestValues", c_uint32,
               [c_int16, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps6000NoOfStreamingValues
        (
            int16_t   handle,
            uint32_t *noOfValues
        ); """
    def_symbol(lib, "ps6000NoOfStreamingValues", c_uint32, [c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps6000GetMaxDownSampleRatio
        (
            int16_t            handle,
            uint32_t           noOfUnaggreatedSamples,
            uint32_t          *maxDownSampleRatio,
            PS6000_RATIO_MODE  downSampleRatioMode,
            uint32_t           segmentIndex
        ); """
    def_symbol(lib, "ps6000GetMaxDownSampleRatio", c_uint32,
               [c_int16, c_uint32, c_void_p, c_int32, c_uint32], doc)

    doc = """ PICO_STATUS ps6000GetValues
        (
            int16_t            handle,
            uint32_t           startIndex,
            uint32_t          *noOfSamples,
            uint32_t           downSampleRatio,
            PS6000_RATIO_MODE  downSampleRatioMode,
            uint32_t           segmentIndex,
            int16_t           *overflow
        ); """
    def_symbol(lib, "ps6000GetValues", c_uint32,
               [c_int16, c_uint32, c_void_p, c_uint32, c_int32, c_uint32, c_void_p],
               doc)

    doc = """ PICO_STATUS ps6000GetValuesBulk
        (
            int16_t            handle,
            uint32_t          *noOfSamples,
            uint32_t           fromSegmentIndex,
            uint32_t           toSegmentIndex,
            uint32_t           downSampleRatio,
            PS6000_RATIO_MODE  downSampleRatioMode,
            int16_t           *overflow
        ); """
    def_symbol(lib, "ps6000GetValuesBulk", c_uint32,
               [c_int16, c_void_p, c_uint32, c_uint32, c_uint32, c_int32, c_void_p],
               doc)

    doc = """ PICO_STATUS ps6000GetValuesAsync
        (
            int16_t            handle,
            uint32_t           startIndex,
            uint32_t           noOfSamples,
            uint32_t           downSampleRatio,
            PS6000_RATIO_MODE  downSampleRatioMode,
            uint32_t           segmentIndex,
            void              *lpDataReady,
            void              *pParameter
        ); """
    def_symbol(lib, "ps6000GetValuesAsync", c_uint32,
               [c_int16, c_uint32, c_uint32, c_uint32, c_int32, c_uint32, c_void_p,
                c_void_p], doc)

    doc = """ PICO_STATUS ps6000GetValuesOverlapped
        (
            int16_t            handle,
            uint32_t           startIndex,
            uint32_t          *noOfSamples,
            uint32_t           downSampleRatio,
            PS6000_RATIO_MODE  downSampleRatioMode,
            uint32_t           segmentIndex,
            int16_t           *overflow
        ); """
    def_symbol(lib, "ps6000GetValuesOverlapped", c_uint32,
               [c_int16, c_uint32, c_void_p, c_uint32, c_int32, c_uint32, c_void_p],
               doc)

    doc = """ PICO_STATUS ps6000GetValuesOverlappedBulk
        (
            int16_t            handle,
            uint32_t           startIndex,
            uint32_t          *noOfSamples,
            uint32_t           downSampleRatio,
            PS6000_RATIO_MODE  downSampleRatioMode,
            uint32_t           fromSegmentIndex,
            uint32_t           toSegmentIndex,
            int16_t           *overflow
        ); """
    def_symbol(lib, "ps6000GetValuesOverlappedBulk", c_uint32,
               [c_int16, c_uint32, c_void_p, c_uint32, c_int32, c_uint32, c_uint32,
                c_void_p], doc)

    doc = """ PICO_STATUS ps6000GetValuesBulkAsyc
        (
            int16_t            handle,
            uint32_t           startIndex,
            uint32_t          *noOfSamples,
            uint32_t           downSampleRatio,
            PS6000_RATIO_MODE  downSampleRatioMode,
            uint32_t           fromSegmentIndex,
            uint32_t           toSegmentIndex,
            int16_t           *overflow
        ); """
    def_symbol(lib, "ps6000GetValuesAsync", c_uint32,
               [c_int16, c_uint32, c_uint32, c_uint32, c_int32, c_uint32, c_uint32,
                c_void_p], doc)

    doc = """ PICO_STATUS ps6000GetNoOfCaptures
        (
            int16_t   handle,
            uint32_t *nCaptures
        ); """
    def_symbol(lib, "ps6000GetNoOfCaptures", c_uint32, [c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps6000GetNoOfProcessedCaptures
        (
            int16_t   handle,
            uint32_t *nProcessedCaptures
        ); """
    def_symbol(lib, "ps6000GetNoOfProcessedCaptures", c_uint32, [c_int16, c_void_p],
               doc)

    """ PICO_STATUS ps6000Stop
        (
            int16_t  handle
        ); """
    def_symbol(lib, "ps6000Stop", c_uint32, [c_int16, ], doc)

    doc = """ PICO_STATUS ps6000SetNoOfCaptures
        (
            int16_t   handle,
            uint32_t  nCaptures
        ); """
    def_symbol(lib, "ps6000SetNoOfCaptures", c_uint32, [c_int16, c_uint32], doc)

    doc = """ PICO_STATUS ps6000SetWaveformLimiter
        (
            int16_t   handle,
            uint32_t  nWaveformsPerSecond
        ); """
    def_symbol(lib, "ps6000SetWaveformLimiter", c_uint32, [c_int16, c_uint32], doc)

    doc = """ PICO_STATUS ps6000EnumerateUnits
        (
            int16_t *count,
            int8_t  *serials,
            int16_t *serialLth
        ); """
    def_symbol(lib, "ps6000EnumerateUnits", c_uint32, [c_void_p, c_void_p, c_void_p],
               doc)

    doc = """ PICO_STATUS ps6000SetExternalClock
        (
            int16_t                    handle,
            PS6000_EXTERNAL_FREQUENCY  frequency,
            int16_t                    threshold
        ); """
    def_symbol(lib, "ps6000SetExternalClock", c_uint32, [c_int16, c_int32, c_int16],
               doc)

    doc = """ PICO_STATUS ps6000PingUnit
        (
            int16_t  handle
        ); """
    def_symbol(lib, "ps6000PingUnit", c_uint32, [c_int16, ], doc)

    doc = """ PICO_STATUS ps6000GetAnalogueOffset
        (
            int16_t          handle,
            PS6000_RANGE     range,
            PS6000_COUPLING  coupling,
            float           *maximumVoltage,
            float           *minimumVoltage
        ); """
    def_symbol(lib, "ps6000GetAnalogueOffset", c_uint32,
               [c_int16, c_int32, c_int32, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps6000GetTriggerInfoBulk
        (
            int16_t        handle,
            PS6000_TRIGGER_INFO    *triggerInfo,
            uint32_t      fromSegmentIndex,
            uint32_t      toSegmentIndex
        ); """
    def_symbol(lib, "ps6000GetTriggerInfoBulk", c_uint16,
               [c_int16, c_void_p, c_uint32, c_uint32], doc)
    # fmt: on

    return lib


_LIB_PS6000 = None


def load_ps6000():
    global _LIB_PS6000
    if _LIB_PS6000 is None:
        _LIB_PS6000 = _load_ps6000_lib()
        return _LIB_PS6000
    else:
        return _LIB_PS6000


###
# Copyright CEA (Commissariat à l'énergie atomique et aux
# énergies alternatives) (2017-2025)
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
###

from ctypes import (
    c_char_p,
    c_double,
    c_float,
    c_int16,
    c_int32,
    c_uint32,
    c_uint64,
    c_void_p,
)
from enum import IntEnum

from .common import def_symbol, load_libps


class PicoInfo(IntEnum):
    """
    Information to be requested to `ps3000aGetUnitInfo`.
    """

    DRIVER_VERSION = 0x00000000
    USB_VERSION = 0x00000001
    HARDWARE_VERSION = 0x00000002
    VARIANT_INFO = 0x00000003
    BATCH_AND_SERIAL = 0x00000004
    CAL_DATE = 0x00000005
    KERNEL_VERSION = 0x00000006
    DIGITAL_HARDWARE_VERSION = 0x00000007
    ANALOGUE_HARDWARE_VERSION = 0x00000008
    FIRMWARE_VERSION_1 = 0x00000009
    FIRMWARE_VERSION_2 = 0x0000000A


def _load_ps3000a_lib(name="ps3000a"):
    lib = load_libps(name)

    # fmt: off

    doc = """ PICO_STATUS ps3000aOpenUnit
        (
            int16_t *handle,
            int8_t  *serial
        ); """
    def_symbol(lib, "ps3000aOpenUnit", c_uint32,
                        [c_void_p, c_char_p], doc)

    doc = """ PICO_STATUS ps3000aOpenUnitAsync
        (
            int16_t *status,
            int8_t  *serial
        ); """
    def_symbol(lib, "ps3000aOpenUnitAsync", c_uint32,
                        [c_void_p, c_char_p], doc)

    doc = """ PICO_STATUS ps3000aOpenUnitProgress
        (
            int16_t *handle,
            int16_t *progressPercent,
            int16_t *complete
        ); """
    def_symbol(lib, "ps3000aOpenUnitProgress",
                        c_uint32, [c_void_p, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetUnitInfo
        (
            int16_t    handle,
            int8_t    *string,
            int16_t    stringLength,
            int16_t   *requiredSize,
            PICO_INFO  info
        ); """
    def_symbol(lib, "ps3000aGetUnitInfo", c_uint32,
                        [c_int16, c_void_p, c_int16, c_void_p, c_uint32],
                        doc)

    doc = """ PICO_STATUS ps3000aFlashLed
        (
            int16_t  handle,
            int16_t  start
        ); """
    def_symbol(lib, "ps3000aFlashLed", c_uint32,
                        [c_int16, c_int16], doc)

    doc = """ PICO_STATUS ps3000aCloseUnit
        (
            int16_t  handle
        ); """
    def_symbol(lib, "ps3000aCloseUnit", c_uint32, [c_int16, ],
                        doc)

    doc = """ PICO_STATUS ps3000aMemorySegments
        (
            int16_t   handle,
            uint32_t  nSegments,
            int32_t  *nMaxSamples
        ); """
    def_symbol(lib, "ps3000aMemorySegments", c_uint32,
                        [c_int16, c_uint32, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aSetChannel
        (
            int16_t          handle,
            PS3000a_CHANNEL  channel,
            int16_t          enabled,
            PS3000a_COUPLING type,
            PS3000a_RANGE    range,
            float            analogOffset
        ); """
    def_symbol(lib, "ps3000aSetChannel", c_uint32,
                        [c_int16, c_int32, c_int16, c_int32, c_int32, c_float],
                        doc)

    doc = """ PICO_STATUS ps3000aSetDigitalPort
        (
            int16_t              handle,
            PS3000a_DIGITAL_PORT port,
            int16_t              enabled,
            int16_t              logicLevel
        ); """
    def_symbol(lib, "ps3000aSetDigitalPort", c_uint32,
                        [c_int16, c_int32, c_int16, c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetBandwidthFilter
        (
            int16_t                    handle,
            PS3000A_CHANNEL            channel,
            PS3000A_BANDWIDTH_LIMITER  bandwidth
        ); """
    def_symbol(lib, "ps3000aSetBandwidthFilter",
                        c_uint32, [c_int16, c_int32, c_int32], doc)

    doc = """ PICO_STATUS ps3000aSetNoOfCaptures
        (
            int16_t   handle,
            uint32_t  nCaptures
        ); """
    def_symbol(lib, "ps3000aSetNoOfCaptures", c_uint32,
                        [c_int16, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aGetTimebase
        (
            int16_t   handle,
            uint32_t  timebase,
            int32_t   noSamples,
            int32_t  *timeIntervalNanoseconds,
            int16_t   oversample,
            int32_t  *maxSamples,
            uint32_t  segmentIndex
        ); """
    def_symbol(lib, "ps3000aGetTimebase", c_uint32,
                        [c_int16, c_uint32, c_int32, c_void_p, c_int16,
                         c_void_p, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aGetTimebase2
        (
            int16_t  handle,
            uint32_t timebase,
            int32_t  noSamples,
            float   *timeIntervalNanoseconds,
            int16_t  oversample,
            int32_t *maxSamples,
            uint32_t segmentIndex
        ); """
    def_symbol(lib, "ps3000aGetTimebase2", c_uint32,
                        [c_int16, c_uint32, c_int32, c_void_p, c_int16,
                         c_void_p, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aSetSigGenArbitrary
        (
            int16_t                     handle,
            int32_t                     offsetVoltage,
            uint32_t                    pkToPk,
            uint32_t                    startDeltaPhase,
            uint32_t                    stopDeltaPhase,
            uint32_t                    deltaPhaseIncrement,
            uint32_t                    dwellCount,
            int16_t                    *arbitraryWaveform,
            int32_t                     arbitraryWaveformSize,
            PS3000A_SWEEP_TYPE          sweepType,
            PS3000A_EXTRA_OPERATIONS    operation,
            PS3000A_INDEX_MODE          indexMode,
            uint32_t                    shots,
            uint32_t                    sweeps,
            PS3000A_SIGGEN_TRIG_TYPE    triggerType,
            PS3000A_SIGGEN_TRIG_SOURCE  triggerSource,
            int16_t                     extInThreshold
        ); """
    def_symbol(lib, "ps3000aSetSigGenArbitrary",
                        c_uint32,
                        [c_int16, c_int32, c_uint32, c_uint32, c_uint32,
                         c_uint32, c_uint32, c_void_p, c_int32, c_int32,
                         c_int32,
                         c_int32, c_uint32, c_uint32, c_int32, c_int32,
                         c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetSigGenBuiltIn
        (
            int16_t                     handle,
            int32_t                     offsetVoltage,
            uint32_t                    pkToPk,
            int16_t                     waveType,
            float                       startFrequency,
            float                       stopFrequency,
            float                       increment,
            float                       dwellTime,
            PS3000A_SWEEP_TYPE          sweepType,
            PS3000A_EXTRA_OPERATIONS    operation,
            uint32_t                    shots,
            uint32_t                    sweeps,
            PS3000A_SIGGEN_TRIG_TYPE    triggerType,
            PS3000A_SIGGEN_TRIG_SOURCE  triggerSource,
            int16_t                     extInThreshold
        ); """
    def_symbol(lib, "ps3000aSetSigGenBuiltIn",
                        c_uint32,
                        [c_int16, c_int32, c_uint32, c_int16, c_float, c_float,
                         c_float, c_float, c_int32, c_int32,
                         c_uint32,
                         c_uint32, c_int32, c_int32, c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetSigGenPropertiesArbitrary
        (
            int16_t                     handle,
            uint32_t                    startDeltaPhase,
            uint32_t                    stopDeltaPhase,
            uint32_t                    deltaPhaseIncrement,
            uint32_t                    dwellCount,
            PS3000A_SWEEP_TYPE          sweepType,
            uint32_t                    shots,
            uint32_t                    sweeps,
            PS3000A_SIGGEN_TRIG_TYPE    triggerType,
            PS3000A_SIGGEN_TRIG_SOURCE  triggerSource,
            int16_t                     extInThreshold
        ); """
    def_symbol(lib,
                        "ps3000aSetSigGenPropertiesArbitrary", c_uint32,
                        [c_int16, c_uint32, c_uint32, c_uint32, c_uint32,
                         c_int32, c_uint32, c_uint32, c_int32, c_int32,
                         c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetSigGenPropertiesBuiltIn
        (
            int16_t                     handle,
            double                      startFrequency,
            double                      stopFrequency,
            double                      increment,
            double                      dwellTime,
            PS3000A_SWEEP_TYPE          sweepType,
            uint32_t                    shots,
            uint32_t                    sweeps,
            PS3000A_SIGGEN_TRIG_TYPE    triggerType,
            PS3000A_SIGGEN_TRIG_SOURCE  triggerSource,
            int16_t                     extInThreshold
        ); """
    def_symbol(lib,
                        "ps3000aSetSigGenPropertiesBuiltIn", c_uint32,
                        [c_int16, c_double, c_double, c_double, c_double,
                         c_int32, c_uint32, c_uint32, c_int32, c_int32,
                         c_int16], doc)

    doc = """ PICO_STATUS ps3000aSigGenFrequencyToPhase
        (
            int16_t             handle,
            double              frequency,
            PS3000A_INDEX_MODE  indexMode,
            uint32_t            bufferLength,
            uint32_t           *phase
        ); """
    def_symbol(lib,
                        "ps3000aSigGenFrequencyToPhase", c_uint32,
                        [c_int16, c_double, c_int32, c_uint32, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aSigGenArbitraryMinMaxValues
        (
            int16_t   handle,
            int16_t  *minArbitraryWaveformValue,
            int16_t  *maxArbitraryWaveformValue,
            uint32_t *minArbitraryWaveformSize,
            uint32_t *maxArbitraryWaveformSize
        ); """
    def_symbol(lib,
                        "ps3000aSigGenArbitraryMinMaxValues", c_uint32,
                        [c_int16, c_void_p, c_void_p, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetMaxEtsValues
        (
            int16_t  handle,
            int16_t *etsCycles,
            int16_t *etsInterleave
        ); """
    def_symbol(lib, "ps3000aGetMaxEtsValues", c_uint32,
                        [c_int16, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aSigGenSoftwareControl
        (
            int16_t  handle,
            int16_t  state
        ); """
    def_symbol(lib,
                        "ps3000aSigGenSoftwareControl", c_uint32,
                        [c_int16, c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetEts
        (
            int16_t           handle,
            PS3000A_ETS_MODE  mode,
            int16_t           etsCycles,
            int16_t           etsInterleave,
            int32_t          *sampleTimePicoseconds
        ); """
    def_symbol(lib, "ps3000aSetEts", c_uint32,
                        [c_int16, c_int32, c_int16, c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aSetSimpleTrigger
        (
            int16_t                      handle,
            int16_t                      enable,
            PS3000A_CHANNEL              source,
            int16_t                      threshold,
            PS3000A_THRESHOLD_DIRECTION  direction,
            uint32_t                     delay,
            int16_t                      autoTrigger_ms
        ); """
    def_symbol(lib, "ps3000aSetSimpleTrigger",
                        c_uint32,
                        [c_int16, c_int16, c_int32, c_int16, c_int32, c_uint32,
                         c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetTriggerDigitalPortProperties
        (
            int16_t                             handle,
            PS3000A_DIGITAL_CHANNEL_DIRECTIONS *directions,
            int16_t                             nDirections
        ); """
    def_symbol(lib,
                        "ps3000aSetTriggerDigitalPortProperties", c_uint32,
                        [c_int16, c_void_p, c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetPulseWidthDigitalPortProperties
        (
            int16_t                             handle,
            PS3000A_DIGITAL_CHANNEL_DIRECTIONS *directions,
            int16_t                             nDirections
        ); """
    def_symbol(lib,
                        "ps3000aSetPulseWidthDigitalPortProperties", c_uint32,
                        [c_int16, c_void_p, c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetTriggerChannelProperties
        (
            int16_t                             handle,
            PS3000A_TRIGGER_CHANNEL_PROPERTIES *channelProperties,
            int16_t                             nChannelProperties,
            int16_t                             auxOutputEnable,
            int32_t                             autoTriggerMilliseconds
        ); """
    def_symbol(lib,
                        "ps3000aSetTriggerChannelProperties", c_uint32,
                        [c_int16, c_void_p, c_int16, c_int16, c_int32], doc)

    doc = """ PICO_STATUS ps3000aSetTriggerChannelConditions
        (
            int16_t                     handle,
            PS3000A_TRIGGER_CONDITIONS *conditions,
            int16_t                     nConditions
        ); """
    def_symbol(lib,
                        "ps3000aSetTriggerChannelConditions", c_uint32,
                        [c_int16, c_void_p, c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetTriggerChannelConditionsV2
        (
            int16_t                        handle,
            PS3000A_TRIGGER_CONDITIONS_V2 *conditions,
            int16_t                        nConditions
        ); """
    def_symbol(lib,
                        "ps3000aSetTriggerChannelConditionsV2", c_uint32,
                        [c_int16, c_void_p, c_int16], doc)

    doc = """ PICO_STATUS ps3000aSetTriggerChannelDirections
        (
            int16_t                      handle,
            PS3000A_THRESHOLD_DIRECTION  channelA,
            PS3000A_THRESHOLD_DIRECTION  channelB,
            PS3000A_THRESHOLD_DIRECTION  channelC,
            PS3000A_THRESHOLD_DIRECTION  channelD,
            PS3000A_THRESHOLD_DIRECTION  ext,
            PS3000A_THRESHOLD_DIRECTION  aux
        ); """
    def_symbol(lib,
                        "ps3000aSetTriggerChannelDirections", c_uint32,
                        [c_int16, c_int32, c_int32, c_int32, c_int32, c_int32,
                         c_int32], doc)

    doc = """ PICO_STATUS ps3000aSetTriggerDelay
        (
            int16_t   handle,
            uint32_t  delay
        ); """
    def_symbol(lib, "ps3000aSetTriggerDelay", c_uint32,
                        [c_int16, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aSetPulseWidthQualifier
        (
            int16_t                      handle,
            PS3000A_PWQ_CONDITIONS      *conditions,
            int16_t                      nConditions,
            PS3000A_THRESHOLD_DIRECTION  direction,
            uint32_t                     lower,
            uint32_t                     upper,
            PS3000A_PULSE_WIDTH_TYPE     type
        ); """
    def_symbol(lib,
                        "ps3000aSetPulseWidthQualifier", c_uint32,
                        [c_int16, c_void_p, c_int16, c_int32, c_uint32,
                         c_uint32, c_int32], doc)

    doc = """ PICO_STATUS ps3000aSetPulseWidthQualifierV2
        (
            int16_t                      handle,
            PS3000A_PWQ_CONDITIONS_V2   *conditions,
            int16_t                      nConditions,
            PS3000A_THRESHOLD_DIRECTION  direction,
            uint32_t                     lower,
            uint32_t                     upper,
            PS3000A_PULSE_WIDTH_TYPE     type
        ); """
    def_symbol(lib,
                        "ps3000aSetPulseWidthQualifierV2", c_uint32,
                        [c_int16, c_void_p, c_int16, c_int32, c_uint32,
                         c_uint32, c_int32], doc)

    doc = """ PICO_STATUS ps3000aIsTriggerOrPulseWidthQualifierEnabled
        (
            int16_t  handle,
            int16_t *triggerEnabled,
            int16_t *pulseWidthQualifierEnabled
        ); """
    def_symbol(lib,
                        "ps3000aIsTriggerOrPulseWidthQualifierEnabled",
                        c_uint32,
                        [c_int16, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetTriggerTimeOffset64
        (
            int16_t             handle,
            int64_t            *time,
            PS3000A_TIME_UNITS *timeUnits,
            uint32_t            segmentIndex
        ); """
    def_symbol(lib,
                        "ps3000aGetTriggerTimeOffset64", c_uint32,
                        [c_int16, c_void_p, c_void_p, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aGetValuesTriggerTimeOffsetBulk64
        (
            int16_t             handle,
            int64_t            *times,
            PS3000A_TIME_UNITS *timeUnits,
            uint32_t            fromSegmentIndex,
            uint32_t            toSegmentIndex
        ); """
    def_symbol(lib,
                        "ps3000aGetValuesTriggerTimeOffsetBulk64", c_uint32,
                        [c_int16, c_void_p, c_void_p, c_uint32, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aGetNoOfCaptures
        (
            int16_t   handle,
            uint32_t *nCaptures
        ); """
    def_symbol(lib, "ps3000aGetNoOfCaptures", c_uint32,
                        [c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetNoOfProcessedCaptures
        (
            int16_t   handle,
            uint32_t *nProcessedCaptures
        ); """
    def_symbol(lib,
                        "ps3000aGetNoOfProcessedCaptures", c_uint32,
                        [c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aSetDataBuffer
        (
            int16_t            handle,
            int32_t            channelOrPort,
            int16_t           *buffer,
            int32_t            bufferLth,
            uint32_t           segmentIndex,
            PS3000a_RATIO_MODE mode
        ); """
    def_symbol(lib, "ps3000aSetDataBuffer", c_uint32,
                        [c_int16, c_int32, c_void_p, c_int32, c_uint32,
                         c_int32], doc)

    doc = """ PICO_STATUS ps3000aSetDataBuffers
        (
            int16_t            handle,
            int32_t            channelOrPort,
            int16_t           *bufferMax,
            int16_t           *bufferMin,
            int32_t            bufferLth,
            uint32_t           segmentIndex,
            PS3000a_RATIO_MODE mode
        ); """
    def_symbol(lib, "ps3000aSetDataBuffers", c_uint32,
                        [c_int16, c_int32, c_void_p, c_void_p, c_int32,
                         c_uint32, c_int32], doc)

    doc = """ PICO_STATUS ps3000aSetEtsTimeBuffer
        (
            int16_t    handle,
            int64_t *buffer,
            int32_t     bufferLth
        ); """
    def_symbol(lib, "ps3000aSetEtsTimeBuffer",
                        c_uint32, [c_int16, c_void_p, c_int32], doc)

    doc = """ PICO_STATUS ps3000aIsReady
        (
            int16_t  handle,
            int16_t *ready
        ); """
    def_symbol(lib, "ps3000aIsReady", c_uint32,
                        [c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aRunBlock
        (
            int16_t            handle,
            int32_t            noOfPreTriggerSamples,
            int32_t            noOfPostTriggerSamples,
            uint32_t           timebase,
            int16_t            oversample,
            int32_t           *timeIndisposedMs,
            uint32_t           segmentIndex,
            ps3000aBlockReady  lpReady,
            void              *pParameter
        ); """
    def_symbol(lib, "ps3000aRunBlock", c_uint32,
                        [c_int16, c_int32, c_int32, c_uint32, c_int16, c_void_p,
                         c_uint32, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aRunStreaming
        (
            int16_t             handle,
            uint32_t            *sampleInterval,
            PS3000A_TIME_UNITS  sampleIntervalTimeUnits,
            uint32_t            maxPreTriggerSamples,
            uint32_t            maxPostPreTriggerSamples,
            int16_t             autoStop,
            uint32_t            downSampleRatio,
            PS3000A_RATIO_MODE  downSampleRatioMode,
            uint32_t            overviewBufferSize
        ); """
    def_symbol(lib, "ps3000aRunStreaming", c_uint32,
                        [c_int16, c_void_p, c_int32, c_uint32, c_uint32,
                         c_int16, c_uint32, c_int32, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aGetStreamingLatestValues
        (
            int16_t                handle,
            ps3000aStreamingReady  lpPs3000aReady,
            void                   *pParameter
        ); """
    def_symbol(lib,
                        "ps3000aGetStreamingLatestValues", c_uint32,
                        [c_int16, c_void_p, c_void_p], doc)

    # doc = """ void *ps3000aStreamingReady
    #     (
    #         int16_t   handle,
    #         int32_t   noOfSamples,
    #         uint32_t  startIndex,
    #         int16_t   overflow,
    #         uint32_t  triggerAt,
    #         int16_t   triggered,
    #         int16_t   autoStop,
    #         void     *pParameter
    #     );
    #     define a python function which accepts the correct arguments, and pass it to the constructor of this type.
    #     """

    # ps3000a.StreamingReadyType = C_CALLBACK_FUNCTION_FACTORY(None,
    #                                                          c_int16,
    #                                                          c_int32,
    #                                                          c_uint32,
    #                                                          c_int16,
    #                                                          c_uint32,
    #                                                          c_int16,
    #                                                          c_int16,
    #                                                          c_void_p)
    #
    # ps3000a.StreamingReadyType.__doc__ = doc

    doc = """ PICO_STATUS ps3000aNoOfStreamingValues
        (
            int16_t   handle,
            uint32_t *noOfValues
        ); """
    def_symbol(lib, "ps3000aNoOfStreamingValues",
                        c_uint32, [c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetMaxDownSampleRatio
    (
      int16_t               handle,
      uint32_t       noOfUnaggreatedSamples,
      uint32_t      *maxDownSampleRatio,
      PS3000A_RATIO_MODE  downSampleRatioMode,
      uint32_t      segmentIndex
    ); """
    def_symbol(lib,
                        "ps3000aGetMaxDownSampleRatio", c_uint32,
                        [c_int16, c_uint32, c_void_p, c_int32, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aGetValues
        (
            int16_t             handle,
            uint32_t            startIndex,
            uint32_t           *noOfSamples,
            uint32_t            downSampleRatio,
            PS3000a_RATIO_MODE  downSampleRatioMode,
            uint32_t            segmentIndex,
            int16_t            *overflow
        ); """
    def_symbol(lib, "ps3000aGetValues", c_uint32,
                        [c_int16, c_uint32, c_void_p, c_uint32, c_int32,
                         c_uint32, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetValuesBulk
        (
            int16_t             handle,
            uint32_t           *noOfSamples,
            uint32_t            fromSegmentIndex,
            uint32_t            toSegmentIndex,
            uint32_t            downSampleRatio,
            PS3000A_RATIO_MODE  downSampleRatioMode,
            int16_t            *overflow
        ); """
    def_symbol(lib, "ps3000aGetValuesBulk", c_uint32,
                        [c_int16, c_void_p, c_uint32, c_uint32, c_uint32,
                         c_int32, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetValuesAsync
        (
            int16_t             handle,
            uint32_t            startIndex,
            uint32_t            noOfSamples,
            uint32_t            downSampleRatio,
            PS3000A_RATIO_MODE  downSampleRatioMode,
            uint32_t            segmentIndex,
            void               *lpDataReady,
            void               *pParameter
        ); """
    def_symbol(lib, "ps3000aGetValuesAsync", c_uint32,
                        [c_int16, c_uint32, c_uint32, c_uint32, c_int32,
                         c_uint32, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetValuesOverlapped
        (
            int16_t             handle,
            uint32_t            startIndex,
            uint32_t           *noOfSamples,
            uint32_t            downSampleRatio,
            PS3000A_RATIO_MODE  downSampleRatioMode,
            uint32_t            segmentIndex,
            int16_t            *overflow
        ); """
    def_symbol(lib, "ps3000aGetValuesOverlapped",
                        c_uint32,
                        [c_int16, c_uint32, c_void_p, c_uint32, c_int32,
                         c_uint32, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetValuesOverlappedBulk
        (
            int16_t             handle,
            uint32_t            startIndex,
            uint32_t           *noOfSamples,
            uint32_t            downSampleRatio,
            PS3000A_RATIO_MODE  downSampleRatioMode,
            uint32_t            fromSegmentIndex,
            uint32_t            toSegmentIndex,
            int16_t            *overflow
        ); """
    def_symbol(lib,"ps3000aGetValuesOverlappedBulk", c_uint32,[c_int16, c_uint32, c_void_p, c_uint32, c_int32, c_uint32, c_uint32, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetTriggerInfoBulk
        (
            int16_t               handle,
            PS3000A_TRIGGER_INFO *triggerInfo,
            uint32_t              fromSegmentIndex,
            uint32_t              toSegmentIndex
        ); """
    def_symbol(lib, "ps3000aGetTriggerInfoBulk", c_uint32, [c_int16, c_void_p, c_uint32, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aStop
        (
            int16_t  handle
        ); """
    def_symbol(lib, "ps3000aStop", c_uint32, [c_int16, ], doc)

    doc = """ PICO_STATUS ps3000aHoldOff
        (
            int16_t               handle,
            uint64_t              holdoff,
            PS3000A_HOLDOFF_TYPE  type
        ); """
    def_symbol(lib, "ps3000aHoldOff", c_uint32, [c_int16, c_uint64, c_int32], doc)

    doc = """ PICO_STATUS ps3000aGetChannelInformation
        (
            int16_t               handle,
            PS3000A_CHANNEL_INFO  info,
            int32_t               probe,
            int32_t              *ranges,
            int32_t              *length,
            int32_t               channels
        ); """
    def_symbol(lib,"ps3000aGetChannelInformation", c_uint32,[c_int16, c_int32, c_int32, c_void_p, c_void_p, c_int32], doc)

    doc = """ PICO_STATUS ps3000aEnumerateUnits
        (
            int16_t *count,
            int8_t  *serials,
            int16_t *serialLth
        ); """
    def_symbol(lib, "ps3000aEnumerateUnits", c_uint32, [c_void_p, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aPingUnit
        (
            int16_t  handle
        ); """
    def_symbol(lib, "ps3000aPingUnit", c_uint32, [c_int16, ], doc)

    doc = """ PICO_STATUS ps3000aMaximumValue
        (
            int16_t  handle,
            int16_t *value
        ); """
    def_symbol(lib, "ps3000aMaximumValue", c_uint32,[c_int16, c_void_p], doc)

    doc = """" PICO_STATUS ps3000aMinimumValue
        (
            int16_t  handle,
            int16_t *value
        ); """
    def_symbol(lib, "ps3000aMinimumValue", c_uint32,[c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetAnalogueOffset
        (
            int16_t           handle,
            PS3000A_RANGE     range,
            PS3000A_COUPLING  coupling,
            float            *maximumVoltage,
            float            *minimumVoltage
        ); """
    def_symbol(lib, "ps3000aGetAnalogueOffset", c_uint32, [c_int16, c_int32, c_int32, c_void_p, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aGetMaxSegments
        (
            int16_t   handle,
            uint32_t *maxSegments
        ); """
    def_symbol(lib, "ps3000aGetMaxSegments", c_uint32,[c_int16, c_void_p], doc)

    doc = """ PICO_STATUS ps3000aChangePowerSource
        (
            int16_t     handle,
            PICO_STATUS powerState
        ); """
    def_symbol(lib, "ps3000aChangePowerSource", c_uint32, [c_int16, c_uint32], doc)

    doc = """ PICO_STATUS ps3000aCurrentPowerSource
        (
            int16_t handle
        ); """
    def_symbol(lib, "ps3000aCurrentPowerSource", c_uint32, [c_int16, c_uint32], doc)
    # fmt: on
    return lib


_LIB_PS3000 = None


def load_ps3000a():
    global _LIB_PS3000
    if _LIB_PS3000 is None:
        _LIB_PS3000 = _load_ps3000a_lib()
        return _LIB_PS3000
    else:
        return _LIB_PS3000